##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
import sys
import os
from unittest.mock import patch

sys.path.append(os.path.abspath('..'))  # current folder is ~/tests\
from idaes.surrogate.pysmo.radial_basis_function import (
    RadialBasisFunctions, FeatureScaling
)
import numpy as np
import pandas as pd
from scipy.spatial import distance
import pytest


class TestFeatureScaling:
    test_data_1d = [[x] for x in range(10)]
    test_data_2d = [[x, (x + 1) ** 2] for x in range(10)]
    test_data_3d = [[x, x + 10, (x + 1) ** 2 + x + 10] for x in range(10)]
    test_data_3d_constant = [[x, 10, (x + 1) ** 2 + 10] for x in range(10)]

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_data_scaling_minmax_01(self, array_type):
        input_array = array_type(self.test_data_1d)
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        expected_output_3 = np.array([[9]])
        expected_output_2 = np.array([[0]])
        expected_output_1 = np.array((input_array - expected_output_2) / (expected_output_3 - expected_output_2))
        np.testing.assert_array_equal(output_3, expected_output_3)
        np.testing.assert_array_equal(output_2, expected_output_2)
        np.testing.assert_array_equal(output_1, expected_output_1.reshape(10, 1))

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_data_scaling_minmax_02(self, array_type):
        input_array = array_type(self.test_data_2d)
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        expected_output_3 = np.array([[9, 100]])
        expected_output_2 = np.array([[0, 1]])
        expected_output_1 = np.array((input_array - expected_output_2) / (expected_output_3 - expected_output_2))
        np.testing.assert_array_equal(output_3, expected_output_3)
        np.testing.assert_array_equal(output_2, expected_output_2)
        np.testing.assert_array_equal(output_1, expected_output_1)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_data_scaling_minmax_03(self, array_type):
        input_array = array_type(self.test_data_3d)
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        expected_output_3 = np.array([[9, 19, 119]])
        expected_output_2 = np.array([[0, 10, 11]])
        expected_output_1 = np.array((input_array - expected_output_2) / (expected_output_3 - expected_output_2))
        np.testing.assert_array_equal(output_3, expected_output_3)
        np.testing.assert_array_equal(output_2, expected_output_2)
        np.testing.assert_array_equal(output_1, expected_output_1)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_data_scaling_minmax_04(self, array_type):
        input_array = array_type(self.test_data_3d_constant)
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        expected_output_3 = np.array([[9, 10, 110]])
        expected_output_2 = np.array([[0, 10, 11]])
        scale = expected_output_3 - expected_output_2
        scale[scale == 0.0] = 1.0
        expected_output_1 = (input_array - expected_output_2) / scale
        np.testing.assert_array_equal(output_3, expected_output_3)
        np.testing.assert_array_equal(output_2, expected_output_2)
        np.testing.assert_array_equal(output_1, expected_output_1)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [list])
    def test_data_scaling_minmax_05(self, array_type):
        input_array = array_type(self.test_data_2d)
        with pytest.raises(TypeError):
            FeatureScaling.data_scaling_minmax(input_array)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_data_unscaling_minmax_01(self, array_type):
        input_array = array_type(self.test_data_1d)
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        output_1 = np.array(output_1).reshape(output_1.shape[0], )
        un_output_1 = FeatureScaling.data_unscaling_minmax(output_1, output_2, output_3)
        np.testing.assert_array_equal(un_output_1, np.array(input_array).reshape(10, 1))

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_data_unscaling_minmax_02(self, array_type):
        input_array = array_type(self.test_data_2d)
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        un_output_1 = FeatureScaling.data_unscaling_minmax(output_1, output_2, output_3)
        np.testing.assert_array_equal(un_output_1, input_array)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_data_unscaling_minmax_03(self, array_type):
        input_array = array_type(self.test_data_3d)
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        un_output_1 = FeatureScaling.data_unscaling_minmax(output_1, output_2, output_3)
        np.testing.assert_array_equal(un_output_1, input_array)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_data_unscaling_minmax_04(self, array_type):
        input_array = array_type(self.test_data_3d_constant)
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)
        un_output_1 = FeatureScaling.data_unscaling_minmax(output_1, output_2, output_3)
        np.testing.assert_array_equal(un_output_1, input_array)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_data_unscaling_minmax_05(self, array_type):
        input_array = array_type(self.test_data_2d)
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)

        min_array = np.array([[1]])
        max_array = np.array([[5]])
        with pytest.raises(IndexError):
            FeatureScaling.data_unscaling_minmax(output_1, min_array, max_array)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_data_unscaling_minmax_06(self, array_type):
        input_array = array_type(self.test_data_2d)
        output_1, output_2, output_3 = FeatureScaling.data_scaling_minmax(input_array)

        min_array = np.array([[1, 2, 3]])
        max_array = np.array([[5, 6, 7]])
        with pytest.raises(IndexError):
            FeatureScaling.data_unscaling_minmax(output_1, min_array, max_array)


class TestRadialBasisFunction:
    y = np.array(
        [[i, j, ((i + 1) ** 2) + ((j + 1) ** 2)] for i in np.linspace(0, 10, 21) for j in np.linspace(0, 10, 21)])
    full_data = {'x1': y[:, 0], 'x2': y[:, 1], 'y': y[:, 2]}
    training_data = [[i, j, ((i + 1) ** 2) + ((j + 1) ** 2)] for i in np.linspace(0, 10, 5) for j in
                     np.linspace(0, 10, 5)]
    test_data = [[i, (i + 1) ** 2] for i in range(10)]
    test_data_large = [[i, (i + 1) ** 2] for i in range(200)]
    test_data_1d = [[(i + 1) ** 2] for i in range(10)]
    test_data_3d = [[i, (i + 1) ** 2, (i + 2) ** 2] for i in range(10)]
    sample_points = [[i, (i + 1) ** 2] for i in range(8)]
    sample_points_large = [[i, (i + 1) ** 2] for i in range(100)]
    sample_points_1d = [[(i + 1) ** 2] for i in range(8)]
    sample_points_3d = [[i, (i + 1) ** 2, (i + 2) ** 2] for i in range(8)]

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test__init__01(self, array_type):
        input_array = array_type(self.test_data)
        RbfClass = RadialBasisFunctions(input_array, basis_function=None, solution_method=None, regularization=None)
        assert RbfClass.solution_method == 'algebraic'
        assert RbfClass.basis_function == 'gaussian'
        assert RbfClass.regularization == True

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test__init__02(self, array_type):
        input_array = array_type(self.test_data)
        RbfClass = RadialBasisFunctions(input_array, basis_function='LineaR', solution_method='PyoMo',
                                        regularization=False)
        assert RbfClass.solution_method == 'pyomo'
        assert RbfClass.basis_function == 'linear'
        assert RbfClass.regularization == False

    @pytest.mark.unit
    def test__init__03(self):
        with pytest.raises(Exception):
            RbfClass = RadialBasisFunctions([1, 2, 3, 4], basis_function='LineaR', solution_method='PyoMo',
                                            regularization=False)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test__init__04(self, array_type):
        with pytest.raises(Exception):
            input_array = array_type(self.test_data)
            RbfClass = RadialBasisFunctions(input_array, basis_function=None, solution_method=1, regularization=None)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test__init__05(self, array_type):
        with pytest.raises(Exception):
            input_array = array_type(self.test_data)
            RbfClass = RadialBasisFunctions(input_array, basis_function=None, solution_method='idaes',
                                            regularization=None)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test__init__06(self, array_type):
        with pytest.raises(Exception):
            input_array = array_type(self.test_data)
            RbfClass = RadialBasisFunctions(input_array, basis_function=1, solution_method=None, regularization=None)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test__init__07(self, array_type):
        with pytest.raises(Exception):
            input_array = array_type(self.test_data)
            RbfClass = RadialBasisFunctions(input_array, basis_function='idaes', solution_method=None,
                                            regularization=None)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test__init__08(self, array_type):
        with pytest.raises(Exception):
            input_array = array_type(self.test_data)
            RbfClass = RadialBasisFunctions(input_array, basis_function=None, solution_method=None, regularization=1)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test__init__09(self, array_type):
        with pytest.raises(Exception):
            input_array = array_type(self.test_data)
            RbfClass = RadialBasisFunctions(input_array, basis_function='LineaR', solution_method='PyoMo',
                                            regularization=False, overwrite=1)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test__init__10(self, array_type):
        with pytest.raises(Exception):
            input_array = array_type(self.test_data)
            RbfClass = RadialBasisFunctions(input_array, basis_function='LineaR', solution_method='PyoMo',
                                            regularization=False, fname='solution.pkl')

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test__init__11(self, array_type):
        with pytest.raises(Exception):
            input_array = array_type(self.test_data)
            RbfClass = RadialBasisFunctions(input_array, basis_function='LineaR', solution_method='PyoMo',
                                            regularization=False, fname=1)

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test__init__12(self, array_type):
        file_name = 'test_filename.pickle'
        input_array = array_type(self.test_data)
        RbfClass1 = RadialBasisFunctions(input_array, basis_function='LineaR', solution_method='PyoMo',
                                         regularization=False, fname=file_name, overwrite=True)
        p = RbfClass1.get_feature_vector()
        results = RbfClass1.rbf_training()
        RbfClass2 = RadialBasisFunctions(input_array, basis_function='LineaR', solution_method='PyoMo',
                                         regularization=False, fname=file_name, overwrite=True)
        assert RbfClass1.filename == RbfClass2.filename

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test__init__14(self, array_type):
        input_array = array_type(self.test_data)
        file_name1 = 'test_filename1.pickle'
        file_name2 = 'test_filename2.pickle'
        RbfClass1 = RadialBasisFunctions(input_array, basis_function='LineaR', solution_method='PyoMo',
                                         regularization=False, fname=file_name1, overwrite=True)
        p = RbfClass1.get_feature_vector()
        RbfClass1.training()
        RbfClass2 = RadialBasisFunctions(input_array, basis_function='LineaR', solution_method='PyoMo',
                                         regularization=False, fname=file_name2, overwrite=True)
        assert RbfClass1.filename == file_name1
        assert RbfClass2.filename == file_name2

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_r2_distance(self, array_type):
        input_array = array_type(self.training_data)
        u = np.array([[0.1, 0.9]])
        data_feed = RadialBasisFunctions(input_array)
        output = data_feed.r2_distance(u)
        scaled = FeatureScaling.data_scaling_minmax(input_array)
        scaled = scaled[0]
        scaled_x = np.array(scaled)[:, :-1]
        expected_output = np.sqrt(np.sum(np.square(scaled_x - u), axis=1))
        np.testing.assert_almost_equal(expected_output, output, decimal=6)

    @pytest.mark.unit
    def test_gaussian_basis_transformation(self):
        d_vec = np.array([[0, 0], [5e-6, 7e-6], [0.005, 0.007], [0.05, 0.07], [0.5, 0.7], [5, 7], [50, 70]])
        shape_list = [0.001, 1, 1000]
        expected_output_1 = np.exp(-1 * ((d_vec * shape_list[0]) ** 2))
        expected_output_2 = np.exp(-1 * ((d_vec * shape_list[1]) ** 2))
        expected_output_3 = np.exp(-1 * ((d_vec * shape_list[2]) ** 2))
        output_1 = RadialBasisFunctions.gaussian_basis_transformation(d_vec, shape_list[0])
        output_2 = RadialBasisFunctions.gaussian_basis_transformation(d_vec, shape_list[1])
        output_3 = RadialBasisFunctions.gaussian_basis_transformation(d_vec, shape_list[2])
        np.testing.assert_array_equal(expected_output_1, output_1)
        np.testing.assert_array_equal(expected_output_2, output_2)
        np.testing.assert_array_equal(expected_output_3, output_3)

    @pytest.mark.unit
    def test_linear_transformation(self):
        d_vec = np.array([[0, 0], [5e-6, 7e-6], [0.005, 0.007], [0.05, 0.07], [0.5, 0.7], [5, 7], [50, 70]])
        output_1 = RadialBasisFunctions.linear_transformation(d_vec)
        np.testing.assert_array_equal(d_vec, output_1)

    @pytest.mark.unit
    def test_cubic_transformation(self):
        d_vec = np.array([[0, 0], [5e-6, 7e-6], [0.005, 0.007], [0.05, 0.07], [0.5, 0.7], [5, 7], [50, 70]])
        expected_output = d_vec ** 3
        output = RadialBasisFunctions.cubic_transformation(d_vec)
        np.testing.assert_array_equal(expected_output, output)

    @pytest.mark.unit
    def test_multiquadric_basis_transformation(self):
        d_vec = np.array([[0, 0], [5e-6, 7e-6], [0.005, 0.007], [0.05, 0.07], [0.5, 0.7], [5, 7], [50, 70]])
        shape_list = [0.001, 1, 1000]
        expected_output_1 = np.sqrt(((d_vec * shape_list[0]) ** 2) + 1)
        expected_output_2 = np.sqrt(((d_vec * shape_list[1]) ** 2) + 1)
        expected_output_3 = np.sqrt(((d_vec * shape_list[2]) ** 2) + 1)
        output_1 = RadialBasisFunctions.multiquadric_basis_transformation(d_vec, shape_list[0])
        output_2 = RadialBasisFunctions.multiquadric_basis_transformation(d_vec, shape_list[1])
        output_3 = RadialBasisFunctions.multiquadric_basis_transformation(d_vec, shape_list[2])
        np.testing.assert_array_equal(expected_output_1, output_1)
        np.testing.assert_array_equal(expected_output_2, output_2)
        np.testing.assert_array_equal(expected_output_3, output_3)

    @pytest.mark.unit
    def test_inverse_multiquadric_basis_transformation(self):
        d_vec = np.array([[0, 0], [5e-6, 7e-6], [0.005, 0.007], [0.05, 0.07], [0.5, 0.7], [5, 7], [50, 70]])
        shape_list = [0.001, 1, 1000]
        expected_output_1 = 1 / np.sqrt(((d_vec * shape_list[0]) ** 2) + 1)
        expected_output_2 = 1 / np.sqrt(((d_vec * shape_list[1]) ** 2) + 1)
        expected_output_3 = 1 / np.sqrt(((d_vec * shape_list[2]) ** 2) + 1)
        output_1 = RadialBasisFunctions.inverse_multiquadric_basis_transformation(d_vec, shape_list[0])
        output_2 = RadialBasisFunctions.inverse_multiquadric_basis_transformation(d_vec, shape_list[1])
        output_3 = RadialBasisFunctions.inverse_multiquadric_basis_transformation(d_vec, shape_list[2])
        np.testing.assert_array_equal(expected_output_1, output_1)
        np.testing.assert_array_equal(expected_output_2, output_2)
        np.testing.assert_array_equal(expected_output_3, output_3)

    @pytest.mark.unit
    def test_thin_plate_spline_transformation(self):
        d_vec = np.array([[5e-6, 7e-6], [0.005, 0.007], [0.05, 0.07], [0.5, 0.7], [5, 7], [50, 70], [50, np.NaN]])
        expected_output = np.nan_to_num(d_vec ** 2 * np.log(d_vec))
        output = RadialBasisFunctions.thin_plate_spline_transformation(d_vec)
        np.testing.assert_array_equal(expected_output, output)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_basis_generation(self, array_type):
        input_array = array_type(self.training_data)
        scaled = FeatureScaling.data_scaling_minmax(input_array[0:3])
        scaled = scaled[0]
        scaled_x = np.array(scaled)[:, :-1]
        distance_array = distance.cdist(scaled_x, scaled_x, 'euclidean')

        # Linear
        data_feed_01 = RadialBasisFunctions(input_array[0:3], basis_function='linear')
        expected_output_1 = distance_array
        output_1 = data_feed_01.basis_generation(2)
        np.testing.assert_array_equal(expected_output_1, output_1)

        # Cubic
        data_feed_02 = RadialBasisFunctions(input_array[0:3], basis_function='cubic')
        expected_output_2 = distance_array ** 3
        output_2 = data_feed_02.basis_generation(2)
        np.testing.assert_array_equal(expected_output_2, output_2)

        # # Spline
        data_feed_03 = RadialBasisFunctions(input_array[0:3], basis_function='spline')
        expected_output_3 = np.nan_to_num(distance_array ** 2 * np.log(distance_array))
        output_3 = data_feed_03.basis_generation(2)
        np.testing.assert_array_equal(expected_output_3, output_3)

        # # Gaussian
        data_feed_04 = RadialBasisFunctions(input_array[0:3], basis_function='gaussian')
        shape_value = 2
        expected_output_4 = np.exp(-1 * ((distance_array * shape_value) ** 2))
        output_4 = data_feed_04.basis_generation(shape_value)
        np.testing.assert_array_equal(expected_output_4, output_4)

        # # Multiquadric
        data_feed_05 = RadialBasisFunctions(input_array[0:3], basis_function='mq')
        shape_value = 2
        expected_output_5 = np.sqrt(((distance_array * shape_value) ** 2) + 1)
        output_5 = data_feed_05.basis_generation(shape_value)
        np.testing.assert_array_equal(expected_output_5, output_5)

        # # Inverse multiquadric
        data_feed_06 = RadialBasisFunctions(input_array[0:3], basis_function='imq')
        shape_value = 2
        expected_output_6 = 1 / np.sqrt(((distance_array * shape_value) ** 2) + 1)
        output_6 = data_feed_06.basis_generation(shape_value)
        np.testing.assert_array_equal(expected_output_6, output_6)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_cost_function_01(self, array_type):
        input_array = array_type(self.training_data)
        x = input_array[:, :-1]
        y = input_array[:, -1]
        x_data_nr = x.shape[0]
        x_data_nc = 6
        x_vector = np.zeros((x_data_nr, x_data_nc))
        x_vector[:, 0] = 1
        x_vector[:, 1] = x[:, 0]
        x_vector[:, 2] = x[:, 1]
        x_vector[:, 3] = x[:, 0] ** 2
        x_vector[:, 4] = x[:, 1] ** 2
        x_vector[:, 5] = x[:, 0] * x[:, 1]
        theta = np.zeros((x_data_nc, 1))
        expected_value = 6613.875
        output_1 = RadialBasisFunctions.cost_function(theta, x_vector, y)
        assert output_1 == expected_value

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_cost_function_02(self, array_type):
        input_array = array_type(self.training_data)
        x = input_array[:, :-1]
        y = input_array[:, -1]
        x_data_nr = x.shape[0]
        x_data_nc = 6
        x_vector = np.zeros((x_data_nr, x_data_nc))
        x_vector[:, 0] = 1
        x_vector[:, 1] = x[:, 0]
        x_vector[:, 2] = x[:, 1]
        x_vector[:, 3] = x[:, 0] ** 2
        x_vector[:, 4] = x[:, 1] ** 2
        x_vector[:, 5] = x[:, 0] * x[:, 1]
        theta = np.array([[4.5], [3], [3], [1], [1], [0]])
        expected_value = 90.625  # Calculated externally as sum(dy^2) / 2m
        output_1 = RadialBasisFunctions.cost_function(theta, x_vector, y)
        assert output_1 == expected_value

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_cost_function_03(self, array_type):
        input_array = array_type(self.training_data)
        x = input_array[:, :-1]
        y = input_array[:, -1]
        x_data_nr = x.shape[0]
        x_data_nc = 6
        x_vector = np.zeros((x_data_nr, x_data_nc))
        x_vector[:, 0] = 1
        x_vector[:, 1] = x[:, 0]
        x_vector[:, 2] = x[:, 1]
        x_vector[:, 3] = x[:, 0] ** 2
        x_vector[:, 4] = x[:, 1] ** 2
        x_vector[:, 5] = x[:, 0] * x[:, 1]
        theta = np.array([[2], [2], [2], [1], [1], [0]])
        expected_value = 0
        output_1 = RadialBasisFunctions.cost_function(theta, x_vector, y)
        assert output_1 == expected_value

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_gradient_function_01(self, array_type):
        input_array = array_type(self.training_data)
        x = input_array[:, :-1]
        y = input_array[:, -1]
        x_data_nr = x.shape[0]
        x_data_nc = 6
        x_vector = np.zeros((x_data_nr, x_data_nc))
        x_vector[:, 0] = 1
        x_vector[:, 1] = x[:, 0]
        x_vector[:, 2] = x[:, 1]
        x_vector[:, 3] = x[:, 0] ** 2
        x_vector[:, 4] = x[:, 1] ** 2
        x_vector[:, 5] = x[:, 0] * x[:, 1]
        theta = np.zeros((x_data_nc,))
        expected_value = np.array(
            [[-97], [-635], [-635], [-5246.875], [-5246.875], [-3925]])
        expected_value = expected_value.reshape(expected_value.shape[0], )
        output_1 = RadialBasisFunctions.gradient_function(theta, x_vector, y)
        np.testing.assert_equal(output_1, expected_value)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_gradient_function_02(self, array_type):
        input_array = array_type(self.training_data)
        x = input_array[:, :-1]
        y = input_array[:, -1]
        x_data_nr = x.shape[0]
        x_data_nc = 6
        x_vector = np.zeros((x_data_nr, x_data_nc))
        x_vector[:, 0] = 1
        x_vector[:, 1] = x[:, 0]
        x_vector[:, 2] = x[:, 1]
        x_vector[:, 3] = x[:, 0] ** 2
        x_vector[:, 4] = x[:, 1] ** 2
        x_vector[:, 5] = x[:, 0] * x[:, 1]
        theta = np.array([[4.5], [3], [3], [1], [1], [0]])  # coefficients in (x1 + 1.5)^2 + (x2 + 1.5) ^ 2
        theta = theta.reshape(theta.shape[0], )
        expected_value = np.array(
            [[12.5], [75], [75], [593.75], [593.75], [437.5]])  # Calculated externally: see Excel sheet
        expected_value = expected_value.reshape(expected_value.shape[0], )
        output_1 = RadialBasisFunctions.gradient_function(theta, x_vector, y)
        np.testing.assert_equal(output_1, expected_value)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_gradient_function_03(self, array_type):
        input_array = array_type(self.training_data)
        x = input_array[:, :-1]
        y = input_array[:, -1]
        x_data_nr = x.shape[0]
        x_data_nc = 6
        x_vector = np.zeros((x_data_nr, x_data_nc))
        x_vector[:, 0] = 1
        x_vector[:, 1] = x[:, 0]
        x_vector[:, 2] = x[:, 1]
        x_vector[:, 3] = x[:, 0] ** 2
        x_vector[:, 4] = x[:, 1] ** 2
        x_vector[:, 5] = x[:, 0] * x[:, 1]
        theta = np.array([[2], [2], [2], [1], [1], [0]])  # Actual coefficients in (x1 + 1)^2 + (x2 + 1) ^ 2
        theta = theta.reshape(theta.shape[0], )
        expected_value = np.array([[0], [0], [0], [0], [0], [0]])  # Calculated externally: see Excel sheet
        expected_value = expected_value.reshape(expected_value.shape[0], )
        output_1 = RadialBasisFunctions.gradient_function(theta, x_vector, y)
        np.testing.assert_equal(output_1, expected_value)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_bfgs_parameter_optimization_01(self, array_type):
        input_array = np.array([[0, 1], [1, 4], [2, 9], [3, 16], [4, 25], [5, 36], [6, 49], [7, 64], [8, 81], [9, 100]])
        x = input_array[:, 0]
        y = input_array[:, 1]
        x_vector = np.zeros((x.shape[0], 3))
        x_vector[:, 0] = x[:, ] ** 2
        x_vector[:, 1] = x[:, ]
        x_vector[:, 2] = 1
        expected_value = np.array([[1.], [2.], [1.]]).reshape(3, )
        data_feed = RadialBasisFunctions(array_type(self.test_data), basis_function='linear', solution_method='bfgs')
        output_1 = data_feed.bfgs_parameter_optimization(x_vector, y)
        assert data_feed.solution_method == 'bfgs'
        np.testing.assert_array_equal(expected_value, np.round(output_1, 4))

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type1", [np.array])
    @pytest.mark.parametrize("array_type2", [pd.DataFrame])
    def test_bfgs_parameter_optimization_02(self, array_type1, array_type2):
        input_array = array_type1(self.training_data)
        x = input_array[:, : -1]
        y = input_array[:, -1]
        x_vector = np.zeros((x.shape[0], 6))
        x_vector[:, 0] = x[:, 0] ** 2
        x_vector[:, 1] = x[:, 1] ** 2
        x_vector[:, 2] = x[:, 0]
        x_vector[:, 3] = x[:, 1]
        x_vector[:, 4] = x[:, 1] * x[:, 0]
        x_vector[:, 5] = 1
        expected_value = np.array([[1.], [1.], [2.], [2.], [0.], [2.]]).reshape(6, )
        data_feed = RadialBasisFunctions(array_type2(self.full_data), solution_method='bfgs')
        output_1 = data_feed.bfgs_parameter_optimization(x_vector, y)
        assert data_feed.solution_method == 'bfgs'
        np.testing.assert_array_equal(expected_value, np.round(output_1, 4))

    @pytest.mark.unit
    def test_explicit_linear_algebra_solution_01(self):
        input_array = np.array([[0, 1], [1, 4], [2, 9], [3, 16], [4, 25], [5, 36], [6, 49], [7, 64], [8, 81], [9, 100]])
        x = input_array[:, 0]
        y = input_array[:, 1]
        x_vector = np.zeros((x.shape[0], 3))
        x_vector[:, 0] = x[:, ] ** 2
        x_vector[:, 1] = x[:, ]
        x_vector[:, 2] = 1
        expected_value = np.array([[1.], [2.], [1.]]).reshape(3, )
        output_1 = RadialBasisFunctions.explicit_linear_algebra_solution(x_vector, y)
        np.testing.assert_array_equal(expected_value, np.round(output_1, 4))

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_explicit_linear_algebra_solution_02(self, array_type):
        input_array = array_type(self.training_data)
        x = input_array[:, : -1]
        y = input_array[:, -1]
        x_vector = np.zeros((x.shape[0], 6))
        x_vector[:, 0] = x[:, 0] ** 2
        x_vector[:, 1] = x[:, 1] ** 2
        x_vector[:, 2] = x[:, 0]
        x_vector[:, 3] = x[:, 1]
        x_vector[:, 4] = x[:, 1] * x[:, 0]
        x_vector[:, 5] = 1
        expected_value = np.array([[1.], [1.], [2.], [2.], [0.], [2.]]).reshape(6, )
        output_1 = RadialBasisFunctions.explicit_linear_algebra_solution(x_vector, y)
        np.testing.assert_array_equal(expected_value, np.round(output_1, 4))

    @pytest.mark.unit
    def test_pyomo_optimization_01(self):
        input_array = np.array([[0, 1], [1, 4], [2, 9], [3, 16], [4, 25], [5, 36], [6, 49], [7, 64], [8, 81], [9, 100]])
        x = input_array[:, 0]
        y = input_array[:, 1]
        x_vector = np.zeros((x.shape[0], 3))
        x_vector[:, 0] = x[:, ] ** 2
        x_vector[:, 1] = x[:, ]
        x_vector[:, 2] = 1
        expected_value = np.array([[1.], [2.], [1.]])
        output_1 = RadialBasisFunctions.pyomo_optimization(x_vector, y)
        np.testing.assert_array_equal(expected_value, np.round(output_1, 4))

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_pyomo_optimization_02(self, array_type):
        input_array = array_type(self.training_data)
        x = input_array[:, : -1]
        y = input_array[:, -1]
        x_vector = np.zeros((x.shape[0], 6))
        x_vector[:, 0] = x[:, 0] ** 2
        x_vector[:, 1] = x[:, 1] ** 2
        x_vector[:, 2] = x[:, 0]
        x_vector[:, 3] = x[:, 1]
        x_vector[:, 4] = x[:, 1] * x[:, 0]
        x_vector[:, 5] = 1
        expected_value = np.array([[1.], [1.], [2.], [2.], [0.], [2.]])
        output_1 = RadialBasisFunctions.pyomo_optimization(x_vector, y)
        np.testing.assert_array_equal(expected_value, np.round(output_1, 4))

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_error_calculation_01(self, array_type):
        input_array = array_type(self.training_data)
        x = input_array[:, :-1]
        y = input_array[:, -1].reshape(input_array.shape[0], 1)
        x_data_nr = x.shape[0]
        x_data_nc = 6
        x_vector = np.zeros((x_data_nr, x_data_nc))
        x_vector[:, 0] = 1
        x_vector[:, 1] = x[:, 0]
        x_vector[:, 2] = x[:, 1]
        x_vector[:, 3] = x[:, 0] ** 2
        x_vector[:, 4] = x[:, 1] ** 2
        x_vector[:, 5] = x[:, 0] * x[:, 1]
        theta = np.zeros((x_data_nc, 1))
        expected_value_1 = 2 * 6613.875  # Calculated externally as sum(y^2) / m
        expected_value_2 = expected_value_1 ** 0.5
        output_1, output_2, _ = RadialBasisFunctions.error_calculation(theta, x_vector, y)
        assert output_1 == expected_value_1
        assert output_2 == expected_value_2

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_error_calculation_02(self, array_type):
        input_array = array_type(self.training_data)
        x = input_array[:, :-1]
        y = input_array[:, -1].reshape(input_array.shape[0], 1)
        x_data_nr = x.shape[0]
        x_data_nc = 6
        x_vector = np.zeros((x_data_nr, x_data_nc))
        x_vector[:, 0] = 1
        x_vector[:, 1] = x[:, 0]
        x_vector[:, 2] = x[:, 1]
        x_vector[:, 3] = x[:, 0] ** 2
        x_vector[:, 4] = x[:, 1] ** 2
        x_vector[:, 5] = x[:, 0] * x[:, 1]
        theta = np.array([[4.5], [3], [3], [1], [1], [0]])  # coefficients in (x1 + 1.5)^2 + (x2 + 1.5) ^ 2
        expected_value_1 = 2 * 90.625  # Calculated externally as sum(dy^2) / 2m
        expected_value_2 = expected_value_1 ** 0.5
        output_1, output_2, _ = RadialBasisFunctions.error_calculation(theta, x_vector, y)
        assert output_1 == expected_value_1
        assert output_2 == expected_value_2

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_error_calculation_03(self, array_type):
        input_array = array_type(self.training_data)
        x = input_array[:, :-1]
        y = input_array[:, -1].reshape(input_array.shape[0], 1)
        x_data_nr = x.shape[0]
        x_data_nc = 6
        x_vector = np.zeros((x_data_nr, x_data_nc))
        x_vector[:, 0] = 1
        x_vector[:, 1] = x[:, 0]
        x_vector[:, 2] = x[:, 1]
        x_vector[:, 3] = x[:, 0] ** 2
        x_vector[:, 4] = x[:, 1] ** 2
        x_vector[:, 5] = x[:, 0] * x[:, 1]
        theta = np.array([[2], [2], [2], [1], [1], [0]])  # Actual coefficients in (x1 + 1)^2 + (x2 + 1) ^ 2
        expected_value_1 = 2 * 0  # Value should return zero for exact solution
        expected_value_2 = expected_value_1 ** 0.5
        output_1, output_2, _ = RadialBasisFunctions.error_calculation(theta, x_vector, y)
        assert output_1 == expected_value_1
        assert output_2 == expected_value_2

    @pytest.mark.unit
    def test_r2_calculation_01(self):
        y_actual = np.array([[1], [4], [9], [16], [25], [36], [49], [64], [81], [100]])
        y_pred = y_actual * 1.05
        expected_output = 0.993974359  # Evaluated in Excel
        output = RadialBasisFunctions.r2_calculation(y_actual, y_pred)
        assert round(abs(expected_output - output), 7) == 0

    @pytest.mark.unit
    def test_r2_calculation_02(self):
        y_actual = np.array([[1], [4], [9], [16], [25], [36], [49], [64], [81], [100]])
        y_pred = y_actual * 1.50
        expected_output = 0.3974358974  # Evaluated in Excel
        output = RadialBasisFunctions.r2_calculation(y_actual, y_pred)
        assert round(abs(expected_output - output), 7) == 0

    def mock_basis_generation(self, r):
        return np.ones((self.x_data.shape[0], self.x_data.shape[0]))

    def mock_optimization(self, x, y):
        return 500 * np.ones((x.shape[0], 1))

    @patch.object(RadialBasisFunctions, 'basis_generation', mock_basis_generation)
    @patch.object(RadialBasisFunctions, 'explicit_linear_algebra_solution', mock_optimization)
    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_loo_error_estimation_with_rippa_method_01(self, array_type):
        input_array = array_type(self.training_data)
        reg_param = 0.1
        shape_factor = 1
        expected_x = np.ones((input_array.shape[0], input_array.shape[0])) + (
                    reg_param * np.eye(input_array.shape[0], input_array.shape[0]))
        expected_inverse_x = np.diag(np.linalg.pinv(expected_x))
        expected_radial_weights = 500 * np.ones((input_array.shape[0], 1))
        expected_errors = np.linalg.norm(
            expected_radial_weights / (expected_inverse_x.reshape(expected_inverse_x.shape[0], 1)))

        data_feed = RadialBasisFunctions(input_array, solution_method='algebraic')
        _, output_1, output_2 = data_feed.loo_error_estimation_with_rippa_method(shape_factor, reg_param)
        assert output_1 == np.linalg.cond(expected_x)
        np.testing.assert_array_equal(output_2, expected_errors)

    @patch.object(RadialBasisFunctions, 'basis_generation', mock_basis_generation)
    @patch.object(RadialBasisFunctions, 'pyomo_optimization', mock_optimization)
    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_loo_error_estimation_with_rippa_method_02(self, array_type):
        input_array = array_type(self.training_data)
        reg_param = 0.1
        shape_factor = 1
        expected_x = np.ones((input_array.shape[0], input_array.shape[0])) + (
                    reg_param * np.eye(input_array.shape[0], input_array.shape[0]))
        expected_inverse_x = np.diag(np.linalg.pinv(expected_x))
        expected_radial_weights = 500 * np.ones((input_array.shape[0], 1))
        expected_errors = np.linalg.norm(
            expected_radial_weights / (expected_inverse_x.reshape(expected_inverse_x.shape[0], 1)))

        data_feed = RadialBasisFunctions(input_array, solution_method='pyomo')
        _, output_1, output_2 = data_feed.loo_error_estimation_with_rippa_method(shape_factor, reg_param)
        assert output_1 == np.linalg.cond(expected_x)
        np.testing.assert_array_equal(output_2, expected_errors)

    @patch.object(RadialBasisFunctions, 'basis_generation', mock_basis_generation)
    @patch.object(RadialBasisFunctions, 'bfgs_parameter_optimization', mock_optimization)
    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_loo_error_estimation_with_rippa_method_03(self, array_type):
        input_array = array_type(self.training_data)
        reg_param = 0.1
        shape_factor = 1
        expected_x = np.ones((input_array.shape[0], input_array.shape[0])) + (
                    reg_param * np.eye(input_array.shape[0], input_array.shape[0]))
        expected_inverse_x = np.diag(np.linalg.pinv(expected_x))
        expected_radial_weights = 500 * np.ones((input_array.shape[0], 1))
        expected_errors = np.linalg.norm(
            expected_radial_weights / (expected_inverse_x.reshape(expected_inverse_x.shape[0], 1)))

        data_feed = RadialBasisFunctions(input_array, solution_method='bfgs')
        _, output_1, output_2 = data_feed.loo_error_estimation_with_rippa_method(shape_factor, reg_param)
        assert output_1 == np.linalg.cond(expected_x)
        np.testing.assert_array_equal(output_2, expected_errors)

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_leave_one_out_crossvalidation_01(self, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function=None, solution_method=None, regularization=False)
        r_best, lambda_best, error_best = data_feed.leave_one_out_crossvalidation()
        if (data_feed.basis_function == 'gaussian') or (data_feed.basis_function == 'mq') or (
                data_feed.basis_function.lower() == 'imq'):
            r_set = [0.001, 0.002, 0.005, 0.0075, 0.01, 0.02, 0.05, 0.075, 0.1, 0.2, 0.5, 0.75, 1.0, 2.0, 5.0, 7.5,
                     10.0, 20.0, 50.0, 75.0, 100.0, 200.0, 500.0, 1000.0]
        else:
            r_set = [0]
        if data_feed.regularization is True:
            reg_parameter = [0.00001, 0.00002, 0.00005, 0.000075, 0.0001, 0.0002, 0.0005, 0.00075, 0.001, 0.002, 0.005,
                             0.0075, 0.01, 0.02, 0.05, 0.075, 0.1, 0.2, 0.5, 0.75, 1]
        elif data_feed.regularization is False:
            reg_parameter = [0]
        _, _, expected_errors = data_feed.loo_error_estimation_with_rippa_method(r_best, lambda_best)
        assert (r_best in r_set) == True
        assert (lambda_best in reg_parameter) == True
        assert error_best == expected_errors

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_leave_one_out_crossvalidation_02(self, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function='cubic', solution_method=None,
                                         regularization=False)
        r_best, lambda_best, error_best = data_feed.leave_one_out_crossvalidation()
        if (data_feed.basis_function == 'gaussian') or (data_feed.basis_function == 'mq') or (
                data_feed.basis_function.lower() == 'imq'):
            r_set = [0.001, 0.002, 0.005, 0.0075, 0.01, 0.02, 0.05, 0.075, 0.1, 0.2, 0.5, 0.75, 1.0, 2.0, 5.0, 7.5,
                     10.0, 20.0, 50.0, 75.0, 100.0, 200.0, 500.0, 1000.0]
        else:
            r_set = [0]
        if data_feed.regularization is True:
            reg_parameter = [0.00001, 0.00002, 0.00005, 0.000075, 0.0001, 0.0002, 0.0005, 0.00075, 0.001, 0.002, 0.005,
                             0.0075, 0.01, 0.02, 0.05, 0.075, 0.1, 0.2, 0.5, 0.75, 1]
        elif data_feed.regularization is False:
            reg_parameter = [0]
        _, _, expected_errors = data_feed.loo_error_estimation_with_rippa_method(r_best, lambda_best)
        assert (r_best in r_set) == True
        assert (lambda_best in reg_parameter) == True
        assert error_best == expected_errors

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_leave_one_out_crossvalidation_03(self, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function='linear', solution_method=None,
                                         regularization=False)
        r_best, lambda_best, error_best = data_feed.leave_one_out_crossvalidation()
        if (data_feed.basis_function == 'gaussian') or (data_feed.basis_function == 'mq') or (
                data_feed.basis_function.lower() == 'imq'):
            r_set = [0.001, 0.002, 0.005, 0.0075, 0.01, 0.02, 0.05, 0.075, 0.1, 0.2, 0.5, 0.75, 1.0, 2.0, 5.0, 7.5,
                     10.0, 20.0, 50.0, 75.0, 100.0, 200.0, 500.0, 1000.0]
        else:
            r_set = [0]
        if data_feed.regularization is True:
            reg_parameter = [0.00001, 0.00002, 0.00005, 0.000075, 0.0001, 0.0002, 0.0005, 0.00075, 0.001, 0.002, 0.005,
                             0.0075, 0.01, 0.02, 0.05, 0.075, 0.1, 0.2, 0.5, 0.75, 1]
        elif data_feed.regularization is False:
            reg_parameter = [0]
        _, _, expected_errors = data_feed.loo_error_estimation_with_rippa_method(r_best, lambda_best)
        assert (r_best in r_set) == True
        assert (lambda_best in reg_parameter) == True
        assert error_best == expected_errors

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_leave_one_out_crossvalidation_04(self, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function='spline', solution_method=None,
                                         regularization=False)
        r_best, lambda_best, error_best = data_feed.leave_one_out_crossvalidation()
        if (data_feed.basis_function == 'gaussian') or (data_feed.basis_function == 'mq') or (
                data_feed.basis_function.lower() == 'imq'):
            r_set = [0.001, 0.002, 0.005, 0.0075, 0.01, 0.02, 0.05, 0.075, 0.1, 0.2, 0.5, 0.75, 1.0, 2.0, 5.0, 7.5,
                     10.0, 20.0, 50.0, 75.0, 100.0, 200.0, 500.0, 1000.0]
        else:
            r_set = [0]
        if data_feed.regularization is True:
            reg_parameter = [0.00001, 0.00002, 0.00005, 0.000075, 0.0001, 0.0002, 0.0005, 0.00075, 0.001, 0.002, 0.005,
                             0.0075, 0.01, 0.02, 0.05, 0.075, 0.1, 0.2, 0.5, 0.75, 1]
        elif data_feed.regularization is False:
            reg_parameter = [0]
        _, _, expected_errors = data_feed.loo_error_estimation_with_rippa_method(r_best, lambda_best)
        assert (r_best in r_set) == True
        assert (lambda_best in reg_parameter) == True
        assert error_best == expected_errors

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_leave_one_out_crossvalidation_05(self, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function='gaussian', solution_method=None,
                                         regularization=False)
        r_best, lambda_best, error_best = data_feed.leave_one_out_crossvalidation()
        if (data_feed.basis_function == 'gaussian') or (data_feed.basis_function == 'mq') or (
                data_feed.basis_function.lower() == 'imq'):
            r_set = [0.001, 0.002, 0.005, 0.0075, 0.01, 0.02, 0.05, 0.075, 0.1, 0.2, 0.5, 0.75, 1.0, 2.0, 5.0, 7.5,
                     10.0, 20.0, 50.0, 75.0, 100.0, 200.0, 500.0, 1000.0]
        else:
            r_set = [0]
        if data_feed.regularization is True:
            reg_parameter = [0.00001, 0.00002, 0.00005, 0.000075, 0.0001, 0.0002, 0.0005, 0.00075, 0.001, 0.002, 0.005,
                             0.0075, 0.01, 0.02, 0.05, 0.075, 0.1, 0.2, 0.5, 0.75, 1]
        elif data_feed.regularization is False:
            reg_parameter = [0]
        _, _, expected_errors = data_feed.loo_error_estimation_with_rippa_method(r_best, lambda_best)
        assert (r_best in r_set) == True
        assert (lambda_best in reg_parameter) == True
        assert error_best == expected_errors

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_leave_one_out_crossvalidation_06(self, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function='mq', solution_method=None, regularization=False)
        r_best, lambda_best, error_best = data_feed.leave_one_out_crossvalidation()
        if (data_feed.basis_function == 'gaussian') or (data_feed.basis_function == 'mq') or (
                data_feed.basis_function.lower() == 'imq'):
            r_set = [0.001, 0.002, 0.005, 0.0075, 0.01, 0.02, 0.05, 0.075, 0.1, 0.2, 0.5, 0.75, 1.0, 2.0, 5.0, 7.5,
                     10.0, 20.0, 50.0, 75.0, 100.0, 200.0, 500.0, 1000.0]
        else:
            r_set = [0]
        if data_feed.regularization is True:
            reg_parameter = [0.00001, 0.00002, 0.00005, 0.000075, 0.0001, 0.0002, 0.0005, 0.00075, 0.001, 0.002, 0.005,
                             0.0075, 0.01, 0.02, 0.05, 0.075, 0.1, 0.2, 0.5, 0.75, 1]
        elif data_feed.regularization is False:
            reg_parameter = [0]
        _, _, expected_errors = data_feed.loo_error_estimation_with_rippa_method(r_best, lambda_best)
        assert (r_best in r_set) == True
        assert (lambda_best in reg_parameter) == True
        assert error_best == expected_errors

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_leave_one_out_crossvalidation_07(self, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function='imq', solution_method=None, regularization=False)
        r_best, lambda_best, error_best = data_feed.leave_one_out_crossvalidation()
        if (data_feed.basis_function == 'gaussian') or (data_feed.basis_function == 'mq') or (
                data_feed.basis_function.lower() == 'imq'):
            r_set = [0.001, 0.002, 0.005, 0.0075, 0.01, 0.02, 0.05, 0.075, 0.1, 0.2, 0.5, 0.75, 1.0, 2.0, 5.0, 7.5,
                     10.0, 20.0, 50.0, 75.0, 100.0, 200.0, 500.0, 1000.0]
        else:
            r_set = [0]
        if data_feed.regularization is True:
            reg_parameter = [0.00001, 0.00002, 0.00005, 0.000075, 0.0001, 0.0002, 0.0005, 0.00075, 0.001, 0.002, 0.005,
                             0.0075, 0.01, 0.02, 0.05, 0.075, 0.1, 0.2, 0.5, 0.75, 1]
        elif data_feed.regularization is False:
            reg_parameter = [0]
        _, _, expected_errors = data_feed.loo_error_estimation_with_rippa_method(r_best, lambda_best)
        assert (r_best in r_set) == True
        assert (lambda_best in reg_parameter) == True
        assert error_best == expected_errors

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_leave_one_out_crossvalidation_08(self, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function=None, solution_method='algebraic',
                                         regularization=False)
        r_best, lambda_best, error_best = data_feed.leave_one_out_crossvalidation()
        if (data_feed.basis_function == 'gaussian') or (data_feed.basis_function == 'mq') or (
                data_feed.basis_function.lower() == 'imq'):
            r_set = [0.001, 0.002, 0.005, 0.0075, 0.01, 0.02, 0.05, 0.075, 0.1, 0.2, 0.5, 0.75, 1.0, 2.0, 5.0, 7.5,
                     10.0, 20.0, 50.0, 75.0, 100.0, 200.0, 500.0, 1000.0]
        else:
            r_set = [0]
        if data_feed.regularization is True:
            reg_parameter = [0.00001, 0.00002, 0.00005, 0.000075, 0.0001, 0.0002, 0.0005, 0.00075, 0.001, 0.002, 0.005,
                             0.0075, 0.01, 0.02, 0.05, 0.075, 0.1, 0.2, 0.5, 0.75, 1]
        elif data_feed.regularization is False:
            reg_parameter = [0]
        _, _, expected_errors = data_feed.loo_error_estimation_with_rippa_method(r_best, lambda_best)
        assert (r_best in r_set) == True
        assert (lambda_best in reg_parameter) == True
        assert error_best == expected_errors

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_leave_one_out_crossvalidation_09(self, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function=None, solution_method='BFGS', regularization=False)
        r_best, lambda_best, error_best = data_feed.leave_one_out_crossvalidation()
        if (data_feed.basis_function == 'gaussian') or (data_feed.basis_function == 'mq') or (
                data_feed.basis_function.lower() == 'imq'):
            r_set = [0.001, 0.002, 0.005, 0.0075, 0.01, 0.02, 0.05, 0.075, 0.1, 0.2, 0.5, 0.75, 1.0, 2.0, 5.0, 7.5,
                     10.0, 20.0, 50.0, 75.0, 100.0, 200.0, 500.0, 1000.0]
        else:
            r_set = [0]
        if data_feed.regularization is True:
            reg_parameter = [0.00001, 0.00002, 0.00005, 0.000075, 0.0001, 0.0002, 0.0005, 0.00075, 0.001, 0.002, 0.005,
                             0.0075, 0.01, 0.02, 0.05, 0.075, 0.1, 0.2, 0.5, 0.75, 1]
        elif data_feed.regularization is False:
            reg_parameter = [0]
        _, _, expected_errors = data_feed.loo_error_estimation_with_rippa_method(r_best, lambda_best)
        assert (r_best in r_set) == True
        assert (lambda_best in reg_parameter) == True
        assert error_best == expected_errors

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_leave_one_out_crossvalidation_10(self, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function=None, solution_method='pyomo',
                                         regularization=False)
        r_best, lambda_best, error_best = data_feed.leave_one_out_crossvalidation()
        if (data_feed.basis_function == 'gaussian') or (data_feed.basis_function == 'mq') or (
                data_feed.basis_function.lower() == 'imq'):
            r_set = [0.001, 0.002, 0.005, 0.0075, 0.01, 0.02, 0.05, 0.075, 0.1, 0.2, 0.5, 0.75, 1.0, 2.0, 5.0, 7.5,
                     10.0, 20.0, 50.0, 75.0, 100.0, 200.0, 500.0, 1000.0]
        else:
            r_set = [0]
        if data_feed.regularization is True:
            reg_parameter = [0.00001, 0.00002, 0.00005, 0.000075, 0.0001, 0.0002, 0.0005, 0.00075, 0.001, 0.002, 0.005,
                             0.0075, 0.01, 0.02, 0.05, 0.075, 0.1, 0.2, 0.5, 0.75, 1]
        elif data_feed.regularization is False:
            reg_parameter = [0]
        _, _, expected_errors = data_feed.loo_error_estimation_with_rippa_method(r_best, lambda_best)
        assert (r_best in r_set) == True
        assert (lambda_best in reg_parameter) == True
        assert error_best == expected_errors

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_leave_one_out_crossvalidation_11(self, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function=None, solution_method=None, regularization=True)
        r_best, lambda_best, error_best = data_feed.leave_one_out_crossvalidation()
        if (data_feed.basis_function == 'gaussian') or (data_feed.basis_function == 'mq') or (
                data_feed.basis_function.lower() == 'imq'):
            r_set = [0.001, 0.002, 0.005, 0.0075, 0.01, 0.02, 0.05, 0.075, 0.1, 0.2, 0.5, 0.75, 1.0, 2.0, 5.0, 7.5,
                     10.0, 20.0, 50.0, 75.0, 100.0, 200.0, 500.0, 1000.0]
        else:
            r_set = [0]
        if data_feed.regularization is True:
            reg_parameter = [0.00001, 0.00002, 0.00005, 0.000075, 0.0001, 0.0002, 0.0005, 0.00075, 0.001, 0.002, 0.005,
                             0.0075, 0.01, 0.02, 0.05, 0.075, 0.1, 0.2, 0.5, 0.75, 1]
        elif data_feed.regularization is False:
            reg_parameter = [0]
        _, _, expected_errors = data_feed.loo_error_estimation_with_rippa_method(r_best, lambda_best)
        assert (r_best in r_set) == True
        assert (lambda_best in reg_parameter) == True
        assert error_best == expected_errors

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_rbf_training_01(self, array_type):
        input_array = array_type(self.test_data)
        data_feed = RadialBasisFunctions(input_array, basis_function=None, solution_method='algebraic',
                                         regularization=False)
        results = data_feed.training()
        best_r_value, best_lambda_param, _ = data_feed.leave_one_out_crossvalidation()
        x_transformed = data_feed.basis_generation(best_r_value)
        x_transformed = x_transformed + (best_lambda_param * np.eye(x_transformed.shape[0], x_transformed.shape[1]))
        x_condition_number = np.linalg.cond(x_transformed)
        if data_feed.solution_method == 'algebraic':
            radial_weights = data_feed.explicit_linear_algebra_solution(x_transformed, data_feed.y_data)
        elif data_feed.solution_method == 'pyomo':
            radial_weights = data_feed.pyomo_optimization(x_transformed, data_feed.y_data)
        elif data_feed.solution_method == 'bfgs':
            radial_weights = data_feed.bfgs_parameter_optimization(x_transformed, data_feed.y_data)
        radial_weights = radial_weights.reshape(radial_weights.shape[0], 1)
        training_ss_error, rmse_error, y_training_predictions_scaled = data_feed.error_calculation(radial_weights,
                                                                                                   x_transformed,
                                                                                                   data_feed.y_data)
        r_square = data_feed.r2_calculation(data_feed.y_data, y_training_predictions_scaled)
        y_training_predictions = data_feed.data_min[0, -1] + y_training_predictions_scaled * (
                    data_feed.data_max[0, -1] - data_feed.data_min[0, -1])
        np.testing.assert_array_equal(radial_weights, results.weights)
        np.testing.assert_array_equal(best_r_value, results.sigma)
        np.testing.assert_array_equal(best_lambda_param, results.regularization)
        np.testing.assert_array_equal(data_feed.centres, results.centres)
        np.testing.assert_array_equal(y_training_predictions, results.output_predictions)
        np.testing.assert_array_equal(rmse_error, results.rmse)
        np.testing.assert_array_equal(x_condition_number, results.condition_number)
        np.testing.assert_array_equal(data_feed.regularization, results.regularization)
        np.testing.assert_array_equal(r_square, results.R2)
        assert data_feed.basis_function == results.basis_function
        np.testing.assert_array_equal(data_feed.data_min[:, :-1], results.x_data_min)
        np.testing.assert_array_equal(data_feed.data_max[:, :-1], results.x_data_max)
        np.testing.assert_array_equal(data_feed.data_min[:, -1], results.y_data_min)
        np.testing.assert_array_equal(data_feed.data_max[:, -1], results.y_data_max)
        assert results.solution_status == 'ok'

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_rbf_training_02(self, array_type):
        input_array = array_type(self.test_data)
        data_feed = RadialBasisFunctions(input_array, basis_function=None, solution_method='pyomo',
                                         regularization=False)
        data_feed.training()
        with pytest.warns(Warning):
            results = data_feed.training()
            assert data_feed.solution_status == 'unstable solution'

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_rbf_training_03(self, array_type):
        input_array = array_type(self.test_data)
        data_feed = RadialBasisFunctions(input_array, basis_function=None, solution_method='bfgs', regularization=False)
        data_feed.training()
        with pytest.warns(Warning):
            data_feed.training()
            assert data_feed.solution_status == 'unstable solution'

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_rbf_predict_output_01(self, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function='linear', regularization=False)
        results = data_feed.training()
        x_test = np.array([[0, 7.5]])
        output = data_feed.predict_output(x_test)
        data_minimum = results.x_data_min
        data_maximum = results.x_data_max
        scale = data_maximum - data_minimum
        scale[scale == 0.0] = 1.0
        x_pred_scaled = (x_test - data_minimum) / scale
        x_test = x_pred_scaled.reshape(x_test.shape)
        distance_vec = distance.cdist(x_test, results.centres, 'euclidean')
        expected_output = np.matmul(distance_vec, results.weights)
        expected_output = results.y_data_min + expected_output * (results.y_data_max - results.y_data_min)
        assert expected_output == output

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_rbf_predict_output_02(self, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function='cubic', regularization=False)
        results = data_feed.training()
        x_test = np.array([[0, 7.5]])
        output = data_feed.predict_output(x_test)
        data_minimum = data_feed.x_data_min
        data_maximum = results.x_data_max
        scale = data_maximum - data_minimum
        scale[scale == 0.0] = 1.0
        x_pred_scaled = (x_test - data_minimum) / scale
        x_test = x_pred_scaled.reshape(x_test.shape)
        distance_vec = distance.cdist(x_test, results.centres, 'euclidean')
        expected_output = np.matmul(distance_vec ** 3, results.weights)
        expected_output = results.y_data_min + expected_output * (results.y_data_max - results.y_data_min)
        assert expected_output == output

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_rbf_predict_output_03(self, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function='gaussian', regularization=False)
        results = data_feed.training()
        x_test = np.array([[0, 7.5]])
        output = data_feed.predict_output(x_test)
        data_minimum = results.x_data_min
        data_maximum = results.x_data_max
        scale = data_maximum - data_minimum
        scale[scale == 0.0] = 1.0
        x_pred_scaled = (x_test - data_minimum) / scale
        x_test = x_pred_scaled.reshape(x_test.shape)
        distance_vec = distance.cdist(x_test, results.centres, 'euclidean')
        expected_output = np.matmul(np.exp(-1 * ((distance_vec * results.sigma) ** 2)), results.weights)
        expected_output = results.y_data_min + expected_output * (results.y_data_max - results.y_data_min)
        assert expected_output == output

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_rbf_predict_output_04(self, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function='imq', regularization=False)
        results = data_feed.training()
        x_test = np.array([[0, 7.5]])
        output = data_feed.predict_output(x_test)
        data_minimum = data_feed.x_data_min
        data_maximum = results.x_data_max
        scale = data_maximum - data_minimum
        scale[scale == 0.0] = 1.0
        x_pred_scaled = (x_test - data_minimum) / scale
        x_test = x_pred_scaled.reshape(x_test.shape)
        distance_vec = distance.cdist(x_test, results.centres, 'euclidean')
        expected_output = np.matmul(1 / np.sqrt(((distance_vec * results.sigma) ** 2) + 1), results.weights)
        expected_output = results.y_data_min + expected_output * (results.y_data_max - results.y_data_min)
        assert expected_output == output

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_rbf_predict_output_05(self, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function='mq', regularization=False)
        results = data_feed.training()
        x_test = np.array([[0, 7.5]])
        output = data_feed.predict_output(x_test)
        data_minimum = results.x_data_min
        data_maximum = results.x_data_max
        scale = data_maximum - data_minimum
        scale[scale == 0.0] = 1.0
        x_pred_scaled = (x_test - data_minimum) / scale
        x_test = x_pred_scaled.reshape(x_test.shape)
        distance_vec = distance.cdist(x_test, results.centres, 'euclidean')
        expected_output = np.matmul(np.sqrt(((distance_vec * results.sigma) ** 2) + 1), results.weights)
        expected_output = results.y_data_min + expected_output * (results.y_data_max - results.y_data_min)
        assert expected_output == output

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_rbf_predict_output_06(self, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function='spline', regularization=False)
        results = data_feed.training()
        x_test = np.array([[0, 7.5]])
        output = data_feed.predict_output(x_test)
        data_minimum = results.x_data_min
        data_maximum = results.x_data_max
        scale = data_maximum - data_minimum
        scale[scale == 0.0] = 1.0
        x_pred_scaled = (x_test - data_minimum) / scale
        x_test = x_pred_scaled.reshape(x_test.shape)
        distance_vec = distance.cdist(x_test, results.centres, 'euclidean')
        expected_output = np.matmul(np.nan_to_num(distance_vec ** 2 * np.log(distance_vec)), results.weights)
        expected_output = results.y_data_min + expected_output * (results.y_data_max - results.y_data_min)
        assert expected_output == output

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [pd.DataFrame])
    def test_get_feature_vector_01(self, array_type):
        input_array = array_type(self.full_data)
        data_feed = RadialBasisFunctions(input_array, basis_function='linear')
        output = data_feed.get_feature_vector()
        expected_dict = {'x1': 0, 'x2': 0}
        assert expected_dict == output.extract_values()

    @pytest.mark.unit
    @pytest.mark.parametrize("array_type", [np.array])
    def test_get_feature_vector_02(self, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function='linear')
        output = data_feed.get_feature_vector()
        expected_dict = {0: 0, 1: 0}
        assert expected_dict == output.extract_values()

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_rbf_generate_expression_01(self, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function='linear', solution_method=None,
                                         regularization=False)
        p = data_feed.get_feature_vector()
        data_feed.training()
        lv = []
        for i in p.keys():
            lv.append(p[i])
        rbf_expr = data_feed.generate_expression((lv))

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_rbf_generate_expression_02(self, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function='cubic', solution_method=None,
                                         regularization=False)
        p = data_feed.get_feature_vector()
        data_feed.training()
        lv = []
        for i in p.keys():
            lv.append(p[i])
        rbf_expr = data_feed.generate_expression((lv))

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_rbf_generate_expression_03(self, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function='gaussian', solution_method=None,
                                         regularization=False)
        p = data_feed.get_feature_vector()
        results = data_feed.training()
        lv = []
        for i in p.keys():
            lv.append(p[i])
        rbf_expr = results.generate_expression((lv))

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_rbf_generate_expression_04(self, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function='mq', solution_method=None, regularization=False)
        p = data_feed.get_feature_vector()
        results = data_feed.training()
        lv = []
        for i in p.keys():
            lv.append(p[i])
        rbf_expr = results.generate_expression((lv))

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_rbf_generate_expression_05(self, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function='imq', solution_method=None, regularization=False)
        p = data_feed.get_feature_vector()
        results = data_feed.training()
        lv = []
        for i in p.keys():
            lv.append(p[i])
        rbf_expr = results.generate_expression((lv))

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_rbf_generate_expression_06(self, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function='spline', solution_method=None,
                                         regularization=False)
        p = data_feed.get_feature_vector()
        results = data_feed.training()
        lv = []
        for i in p.keys():
            lv.append(p[i])
        rbf_expr = results.generate_expression((lv))

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_pickle_load01(self, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function='spline', solution_method=None,
                                         regularization=False)
        p = data_feed.get_feature_vector()
        data_feed.training()
        data_feed.pickle_load(data_feed.filename)

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_pickle_load02(self, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function='spline', solution_method=None,
                                         regularization=False)
        p = data_feed.get_feature_vector()
        data_feed.training()
        with pytest.raises(Exception):
            data_feed.pickle_load('file_not_existing.pickle')

    @pytest.mark.unit
    @pytest.fixture(scope='module')
    @patch("matplotlib.pyplot.show")
    @pytest.mark.parametrize("array_type", [np.array, pd.DataFrame])
    def test_parity_residual_plots(self, mock_show, array_type):
        input_array = array_type(self.training_data)
        data_feed = RadialBasisFunctions(input_array, basis_function='spline', solution_method=None,
                                         regularization=False)
        p = data_feed.get_feature_vector()
        data_feed.training()
        data_feed.parity_residual_plots()


if __name__ == '__main__':
    pytest.main()



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
These are tests of the RoundingRegression methodology
Set to solve the Best Subset Selection problem MIP
minimize objective = OLS + complexity_penalty*number_regressors_chosen
Additional hyper-parameters and features will be added over time
"""

from idaes.surrogate.roundingRegression.RoundingRegression import RoundingRegression
import numpy as np
import pytest

@pytest.mark.unit
def test_simple_regression():
    """ Tests the identification of a simple sparse regression"""

    # Create Example Data
    # n = number of data points, p = number of variables, k = true sparsity
    n = 300
    p = 10
    k = 3
    true_variables = np.random.choice([i for i in range(p)], size=k, replace=False)
    true_coefficients = np.zeros(p)
    true_coefficients[true_variables] = 1

    # Predictors
    X = np.random.rand(n, p)

    # Response
    Y = X@true_coefficients

    # Set Regression penalty (range: 0 - 1)
    complexity_penalty_factor = 0.1

    # Create Instance of Surrogate Modeler
    RR = RoundingRegression(X, Y, complexity_penalty_factor)

    # Build Model
    RR.build_model()

    # Solution
    # MIP Objective
    objective = RR.opt_obj

    # Full Coefficient Array
    coefficients = RR.opt_coeffs

    # Regressors are numbered from 0 to p-1
    chosen_variables = RR.opt_regressors

    # Predict
    Y_predict = X@coefficients

    # Print
    print("Objective: ", objective)
    print("True Variables: ", true_variables)
    print("Chosen Variables ", chosen_variables)

    # Check if solution makes sense
    assert len(chosen_variables) == np.count_nonzero(coefficients), \
        "Number of variables chosen ({}) does not equal number of nonzeros in coefficient vector ({})" \
    .format(len(chosen_variables), np.count_nonzero(coefficients))

    assert objective >= 0, "Objective ({}) cannot be negative".format(objective)


##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Common methods used by generic framework

Author: A Lee
"""

from pyomo.environ import units as pyunits

from idaes.core.util.exceptions import ConfigurationError, PropertyPackageError
import idaes.logger as idaeslog

# Set up logger
_log = idaeslog.getLogger(__name__)


class GenericPropertyPackageError(PropertyPackageError):
    # Error message for when a property is called for but no option provided
    def __init__(self, block, prop):
        self.prop = prop
        self.block = block

    def __str__(self):
        return f"Generic Property Package instance {self.block} called for " \
               f"{self.prop}, but was not provided with a method " \
               f"for this property. Please add a method for this property " \
               f"in the property parameter configuration."


def get_method(self, config_arg, comp=None):
    """
    Method to inspect configuration argument and return the user-defined
    construction method associated with it.

    This method checks whether the value provided is a method or a
    module. If the value is a module, it looks in the module for a method
    with the same name as the config argument and returns this. If the
    value is a method, the method is returned. If the value is neither a
    module or a method, an ConfigurationError is raised.

    Args:
        config_arg : the configuration argument to look up

    Returns:
        A callable method or a ConfigurationError
    """
    if comp is None:
        source_block = self.params.config
    else:
        source_block = self.params.get_component(comp).config

    try:
        c_arg = getattr(source_block, config_arg)
    except AttributeError:
        raise AttributeError("{} Generic Property Package called for invalid "
                             "configuration option {}. Please contact the "
                             "developer of the property package."
                             .format(self.name, config_arg))

    if c_arg is None:
        raise GenericPropertyPackageError(self, config_arg)

    # Check to see if c_arg has an attribute with the name of the config_arg
    # If so, assume c_arg is a class or module holding property subclasses
    if hasattr(c_arg, config_arg):
        c_arg = getattr(c_arg, config_arg)

    # Try to get the return_expression method from c_arg
    # Otherwise assume c_arg is the return_expression method
    try:
        mthd = c_arg.return_expression
    except AttributeError:
        mthd = c_arg

    # Call the return_expression method
    if callable(mthd):
        return mthd
    else:
        raise ConfigurationError(
                "{} Generic Property Package received invalid value "
                "for argument {}. Value must be a method, a class with a "
                "method named expression or a module containing one of the "
                "previous.".format(self.name, config_arg))


def get_component_object(self, comp):
    """
    Utility method to get a component object from the property parameter block.
    This code is used frequently throughout the generic property pacakge
    libraries.

    Args:
        comp: name of the component object to be returned.

    Returns:
        Component: Component object with name comp.

    """
    return self.params.get_component(comp)


def get_bounds_from_config(b, state, base_units):
    """
    Method to take a 3- or 4-tuple state definition config argument and return
    tuples for the bounds and default value of the Var object.

    Expects the form (lower, default, upper, units) where units is optional

    Args:
        b - StateBlock on which the state vars are to be constructed
        state - name of state var as a string (to be matched with config dict)
        base_units - base units of state var to be used if conversion required

    Returns:
        bounds - 2-tuple of state var bounds in base units
        default_val - default value of state var in base units
    """
    try:
        var_config = b.params.config.state_bounds[state]
    except (KeyError, TypeError):
        # State definition missing
        return (None, None), None

    if len(var_config) == 4:
        # Units provided, need to convert values
        bounds = (pyunits.convert_value(var_config[0],
                                        from_units=var_config[3],
                                        to_units=base_units),
                  pyunits.convert_value(var_config[2],
                                        from_units=var_config[3],
                                        to_units=base_units))
        default_val = pyunits.convert_value(var_config[1],
                                            from_units=var_config[3],
                                            to_units=base_units)
    else:
        bounds = (var_config[0], var_config[2])
        default_val = var_config[1]

    return bounds, default_val



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for generic property package core code

Author: Andrew Lee
"""
import pytest

# Import Pyomo units
from pyomo.environ import (
    ConcreteModel, Constraint, Expression, SolverStatus,
    TerminationCondition, units as pyunits, value)

# Import IDAES cores
from idaes.core import (
    FlowsheetBlock, VaporPhase, Component, MaterialBalanceType,
    ControlVolume1DBlock)

from idaes.generic_models.properties.core.generic.generic_property import (
        GenericParameterBlock, StateIndex)
from idaes.generic_models.properties.core.state_definitions import FTPx
from idaes.generic_models.properties.core.eos.ideal import Ideal
from idaes.generic_models.properties.core.pure import RPP4
from idaes.generic_models.properties.core.reactions.dh_rxn import \
    constant_dh_rxn
from idaes.generic_models.properties.core.reactions.equilibrium_constant import \
    van_t_hoff
from idaes.generic_models.properties.core.reactions.equilibrium_forms import \
    power_law_equil
from idaes.generic_models.properties.core.generic.generic_reaction import (
        ConcentrationForm)

from idaes.generic_models.unit_models import Heater
from idaes.core.util.model_statistics import degrees_of_freedom

from idaes.core.util import get_solver


configuration = {
    # Specifying components
    "components": {
        'a': {"type": Component,
              "enth_mol_ig_comp": RPP4,
              "parameter_data": {
                  "mw": (78.1136E-3, pyunits.kg/pyunits.mol),
                  "pressure_crit": (48.9e5, pyunits.Pa),
                  "temperature_crit": (562.2, pyunits.K),
                  "cp_mol_ig_comp_coeff": {
                      'A': (-40, pyunits.J/pyunits.mol/pyunits.K),
                      'B': (0, pyunits.J/pyunits.mol/pyunits.K**2),
                      'C': (0, pyunits.J/pyunits.mol/pyunits.K**3),
                      'D': (0, pyunits.J/pyunits.mol/pyunits.K**4)},
                  "enth_mol_form_vap_comp_ref": (
                      0, pyunits.J/pyunits.mol)}},
        'b': {"type": Component,
              "enth_mol_ig_comp": RPP4,
              "parameter_data": {
                  "mw": (78.1136E-3, pyunits.kg/pyunits.mol),
                  "pressure_crit": (48.9e5, pyunits.Pa),
                  "temperature_crit": (562.2, pyunits.K),
                  "cp_mol_ig_comp_coeff": {
                      'A': (-40, pyunits.J/pyunits.mol/pyunits.K),
                      'B': (0, pyunits.J/pyunits.mol/pyunits.K**2),
                      'C': (0, pyunits.J/pyunits.mol/pyunits.K**3),
                      'D': (0, pyunits.J/pyunits.mol/pyunits.K**4)},
                  "enth_mol_form_vap_comp_ref": (
                      0, pyunits.J/pyunits.mol)}},
        'c': {"type": Component,
              "enth_mol_ig_comp": RPP4,
              "parameter_data": {
                  "mw": (78.1136E-3, pyunits.kg/pyunits.mol),
                  "pressure_crit": (48.9e5, pyunits.Pa),
                  "temperature_crit": (562.2, pyunits.K),
                  "cp_mol_ig_comp_coeff": {
                      'A': (-40, pyunits.J/pyunits.mol/pyunits.K),
                      'B': (0, pyunits.J/pyunits.mol/pyunits.K**2),
                      'C': (0, pyunits.J/pyunits.mol/pyunits.K**3),
                      'D': (0, pyunits.J/pyunits.mol/pyunits.K**4)},
                  "enth_mol_form_vap_comp_ref": (
                      0, pyunits.J/pyunits.mol)}},
        'd': {"type": Component,
              "enth_mol_ig_comp": RPP4,
              "parameter_data": {
                  "mw": (78.1136E-3, pyunits.kg/pyunits.mol),
                  "pressure_crit": (48.9e5, pyunits.Pa),
                  "temperature_crit": (562.2, pyunits.K),
                  "cp_mol_ig_comp_coeff": {
                      'A': (-40, pyunits.J/pyunits.mol/pyunits.K),
                      'B': (0, pyunits.J/pyunits.mol/pyunits.K**2),
                      'C': (0, pyunits.J/pyunits.mol/pyunits.K**3),
                      'D': (0, pyunits.J/pyunits.mol/pyunits.K**4)},
                  "enth_mol_form_vap_comp_ref": (
                      0, pyunits.J/pyunits.mol)}}},

    # Specifying phases
    "phases":  {'Vap': {"type": VaporPhase,
                        "equation_of_state": Ideal}},

    # Set base units of measurement
    "base_units": {"time": pyunits.s,
                   "length": pyunits.m,
                   "mass": pyunits.kg,
                   "amount": pyunits.mol,
                   "temperature": pyunits.K},

    # Specifying state definition
    "state_definition": FTPx,
    "state_bounds": {"flow_mol": (0, 100, 1000, pyunits.mol/pyunits.s),
                     "temperature": (273.15, 300, 450, pyunits.K),
                     "pressure": (5e4, 1e5, 1e6, pyunits.Pa)},
    "state_components": StateIndex.true,  # need this for inherent rxn terms in material balances
    "pressure_ref": (1e5, pyunits.Pa),
    "temperature_ref": (300, pyunits.K),
    "inherent_reactions": {
                "e1": {"stoichiometry": {("Vap", "a"): -1,
                                         ("Vap", "b"): 1},
                       "heat_of_reaction": constant_dh_rxn,
                       "equilibrium_constant": van_t_hoff,
                       "equilibrium_form": power_law_equil,
                       "concentration_form": ConcentrationForm.moleFraction,
                       "parameter_data": {
                           "dh_rxn_ref": -20000,
                           "k_eq_ref": 2,
                           "T_eq_ref": 350}}}}


class TestInherentReactions(object):
    @pytest.fixture()
    def frame(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.params = GenericParameterBlock(default=configuration)

        return m

    @pytest.mark.unit
    def test_build(self, frame):
        frame.fs.props = frame.fs.params.build_state_block([1], default={
            "defined_state": False})

        assert isinstance(frame.fs.props[1].inherent_equilibrium_constraint,
                          Constraint)
        assert len(frame.fs.props[1].inherent_equilibrium_constraint) == 1

        assert isinstance(frame.fs.props[1].dh_rxn, Expression)
        assert len(frame.fs.props[1].dh_rxn) == 1

    @pytest.mark.component
    def test_heater_w_inherent_rxns_comp_phase(self, frame):
        frame.fs.H101 = Heater(default={
            "property_package": frame.fs.params,
            "material_balance_type": MaterialBalanceType.componentPhase})

        frame.fs.H101.inlet.flow_mol.fix(100)
        frame.fs.H101.inlet.mole_frac_comp.fix(0.25)
        frame.fs.H101.inlet.temperature.fix(350)
        frame.fs.H101.inlet.pressure.fix(101325)

        frame.fs.H101.heat_duty.fix(0)

        assert(degrees_of_freedom(frame)) == 0

        frame.fs.H101.initialize()

        solver = get_solver()

        results = solver.solve(frame)

        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

        assert value(
            frame.fs.H101.control_volume.properties_out[0].k_eq["e1"]) == 2

        assert (
            value(frame.fs.H101.control_volume.properties_out[0].k_eq["e1"]) ==
            pytest.approx(value(frame.fs.H101.outlet.mole_frac_comp[0, "b"] /
                                frame.fs.H101.outlet.mole_frac_comp[0, "a"]),
                          rel=1e-5))

        assert (value(frame.fs.H101.outlet.mole_frac_comp[0, "a"]) ==
                pytest.approx(1/6, rel=1e-5))
        assert (value(frame.fs.H101.outlet.mole_frac_comp[0, "b"]) ==
                pytest.approx(1/3, rel=1e-5))
        assert (value(frame.fs.H101.outlet.mole_frac_comp[0, "c"]) ==
                pytest.approx(1/4, rel=1e-5))
        assert (value(frame.fs.H101.outlet.mole_frac_comp[0, "d"]) ==
                pytest.approx(1/4, rel=1e-5))

        assert (value(frame.fs.H101.outlet.flow_mol[0]) ==
                pytest.approx(100, rel=1e-5))
        assert (value(frame.fs.H101.outlet.temperature[0]) ==
                pytest.approx(350, rel=1e-5))
        assert (value(frame.fs.H101.outlet.pressure[0]) ==
                pytest.approx(101325, rel=1e-5))

    @pytest.mark.component
    def test_heater_w_inherent_rxns_comp_total(self, frame):
        frame.fs.H101 = Heater(default={
            "property_package": frame.fs.params,
            "material_balance_type": MaterialBalanceType.componentTotal})

        frame.fs.H101.inlet.flow_mol.fix(100)
        frame.fs.H101.inlet.mole_frac_comp.fix(0.25)
        frame.fs.H101.inlet.temperature.fix(350)
        frame.fs.H101.inlet.pressure.fix(101325)

        frame.fs.H101.heat_duty.fix(0)

        assert(degrees_of_freedom(frame)) == 0

        frame.fs.H101.initialize()

        solver = get_solver()

        results = solver.solve(frame)

        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

        assert value(
            frame.fs.H101.control_volume.properties_out[0].k_eq["e1"]) == 2

        assert (
            value(frame.fs.H101.control_volume.properties_out[0].k_eq["e1"]) ==
            pytest.approx(value(frame.fs.H101.outlet.mole_frac_comp[0, "b"] /
                                frame.fs.H101.outlet.mole_frac_comp[0, "a"]),
                          rel=1e-5))

        assert (value(frame.fs.H101.outlet.mole_frac_comp[0, "a"]) ==
                pytest.approx(1/6, rel=1e-5))
        assert (value(frame.fs.H101.outlet.mole_frac_comp[0, "b"]) ==
                pytest.approx(1/3, rel=1e-5))
        assert (value(frame.fs.H101.outlet.mole_frac_comp[0, "c"]) ==
                pytest.approx(1/4, rel=1e-5))
        assert (value(frame.fs.H101.outlet.mole_frac_comp[0, "d"]) ==
                pytest.approx(1/4, rel=1e-5))

        assert (value(frame.fs.H101.outlet.flow_mol[0]) ==
                pytest.approx(100, rel=1e-5))
        assert (value(frame.fs.H101.outlet.temperature[0]) ==
                pytest.approx(350, rel=1e-5))
        assert (value(frame.fs.H101.outlet.pressure[0]) ==
                pytest.approx(101325, rel=1e-5))

    @pytest.mark.component
    def test_CV1D_w_inherent_rxns_comp_phase(self, frame):
        frame.fs.cv = ControlVolume1DBlock(default={
            "property_package": frame.fs.params,
            "transformation_method": "dae.finite_difference",
            "transformation_scheme": "BACKWARD",
            "finite_elements": 2})

        frame.fs.cv.add_geometry()

        frame.fs.cv.add_state_blocks(has_phase_equilibrium=False)

        frame.fs.cv.add_material_balances(
            balance_type=MaterialBalanceType.componentPhase)

        frame.fs.cv.add_energy_balances()
        frame.fs.cv.add_momentum_balances()
        frame.fs.cv.apply_transformation()

        frame.fs.cv.properties[0, 0].flow_mol.fix(100)
        frame.fs.cv.properties[0, 0].mole_frac_comp.fix(0.25)
        frame.fs.cv.properties[0, 0].temperature.fix(350)
        frame.fs.cv.properties[0, 0].pressure.fix(101325)

        frame.fs.cv.area.fix(1)
        frame.fs.cv.length.fix(1)

        assert(degrees_of_freedom(frame)) == 0

        frame.fs.cv.initialize()

        solver = get_solver()

        results = solver.solve(frame, tee=True)

        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

        assert value(
            frame.fs.cv.properties[0, 1].k_eq["e1"]) == 2

        assert (
            value(frame.fs.cv.properties[0, 1].k_eq["e1"]) ==
            pytest.approx(value(
                frame.fs.cv.properties[0, 1].mole_frac_comp["b"] /
                frame.fs.cv.properties[0, 1].mole_frac_comp["a"]),
                          rel=1e-5))

        assert (value(frame.fs.cv.properties[0, 1].mole_frac_comp["a"]) ==
                pytest.approx(1/6, rel=1e-5))
        assert (value(frame.fs.cv.properties[0, 1].mole_frac_comp["b"]) ==
                pytest.approx(1/3, rel=1e-5))
        assert (value(frame.fs.cv.properties[0, 1].mole_frac_comp["c"]) ==
                pytest.approx(1/4, rel=1e-5))
        assert (value(frame.fs.cv.properties[0, 1].mole_frac_comp["d"]) ==
                pytest.approx(1/4, rel=1e-5))

        assert (value(frame.fs.cv.properties[0, 1].flow_mol) ==
                pytest.approx(100, rel=1e-5))
        assert (value(frame.fs.cv.properties[0, 1].temperature) ==
                pytest.approx(350, rel=1e-5))
        assert (value(frame.fs.cv.properties[0, 1].pressure) ==
                pytest.approx(101325, rel=1e-5))

    @pytest.mark.component
    def test_CV1D_w_inherent_rxns_comp_total(self, frame):
        frame.fs.cv = ControlVolume1DBlock(default={
            "property_package": frame.fs.params,
            "transformation_method": "dae.finite_difference",
            "transformation_scheme": "BACKWARD",
            "finite_elements": 2})

        frame.fs.cv.add_geometry()

        frame.fs.cv.add_state_blocks(has_phase_equilibrium=False)

        frame.fs.cv.add_material_balances(
            balance_type=MaterialBalanceType.componentTotal)

        frame.fs.cv.add_energy_balances()
        frame.fs.cv.add_momentum_balances()
        frame.fs.cv.apply_transformation()

        frame.fs.cv.properties[0, 0].flow_mol.fix(100)
        frame.fs.cv.properties[0, 0].mole_frac_comp.fix(0.25)
        frame.fs.cv.properties[0, 0].temperature.fix(350)
        frame.fs.cv.properties[0, 0].pressure.fix(101325)

        frame.fs.cv.area.fix(1)
        frame.fs.cv.length.fix(1)

        assert(degrees_of_freedom(frame)) == 0

        frame.fs.cv.initialize()

        solver = get_solver()

        results = solver.solve(frame)

        assert results.solver.termination_condition == \
            TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

        assert value(
            frame.fs.cv.properties[0, 1].k_eq["e1"]) == 2

        assert (
            value(frame.fs.cv.properties[0, 1].k_eq["e1"]) ==
            pytest.approx(value(
                frame.fs.cv.properties[0, 1].mole_frac_comp["b"] /
                frame.fs.cv.properties[0, 1].mole_frac_comp["a"]),
                          rel=1e-5))

        assert (value(frame.fs.cv.properties[0, 1].mole_frac_comp["a"]) ==
                pytest.approx(1/6, rel=1e-5))
        assert (value(frame.fs.cv.properties[0, 1].mole_frac_comp["b"]) ==
                pytest.approx(1/3, rel=1e-5))
        assert (value(frame.fs.cv.properties[0, 1].mole_frac_comp["c"]) ==
                pytest.approx(1/4, rel=1e-5))
        assert (value(frame.fs.cv.properties[0, 1].mole_frac_comp["d"]) ==
                pytest.approx(1/4, rel=1e-5))

        assert (value(frame.fs.cv.properties[0, 1].flow_mol) ==
                pytest.approx(100, rel=1e-5))
        assert (value(frame.fs.cv.properties[0, 1].temperature) ==
                pytest.approx(350, rel=1e-5))
        assert (value(frame.fs.cv.properties[0, 1].pressure) ==
                pytest.approx(101325, rel=1e-5))



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
import numpy as np

from ..geometry import Parallelepiped, RectPrism
from ..atom import Atom


def readPointsAndAtomsFromPOSCAR(filename, ImpliedElems=None):
    Pts = []
    Atoms = []
    BBox = Parallelepiped.fromPOSCAR(filename)
    blnAtomToAdd = False  # Used to track
    Elem = None  # Used to store most recent element in the new format
    with open(filename, 'r') as infile:
        CommentLine = infile.readline()
        GSLine = infile.readline().split()
        GS = float(GSLine[0])
        VxLine = infile.readline()
        VyLine = infile.readline()
        VzLine = infile.readline()
        ElementNamesOrCountsLine = infile.readline().split()
        if ElementNamesOrCountsLine[0].isdigit():
            if ImpliedElems is not None:
                Elems = ImpliedElems
                ElemCountsLine = ElementNamesOrCountsLine
            else:
                raise ValueError('Tried to read a POSCAR without explicit elements provided')
        else:
            assert (ElementNamesOrCountsLine[0].isalpha())
            ElemsLine = ElementNamesOrCountsLine
            Elems = [Atom(Elem) for Elem in ElemsLine]
            ElemCountsLine = infile.readline().split()
        ElemCounts = [int(ElemCount) for ElemCount in ElemCountsLine]
        CartesianOrDirectLine = infile.readline().split()
        CorDFlag = CartesianOrDirectLine[0][0].lower()
        blnIsCartesian = (CorDFlag == 'c' or CorDFlag == 'k')
        for t, Elem in enumerate(Elems):
            for i in range(ElemCounts[t]):
                XYZLine = infile.readline().split()
                if blnIsCartesian:
                    P = GS * np.array([float(XYZLine[0]), float(XYZLine[1]), float(XYZLine[2])], dtype=float)
                else:
                    # NOTE: The global scaling GS has already been taken in the BBox
                    #       so no need to multiply it here
                    Pfrac = np.array([float(XYZLine[0]), float(XYZLine[1]), float(XYZLine[2])], dtype=float)
                    P = np.inner(np.array([BBox.Vx, BBox.Vy, BBox.Vz]).T, Pfrac)
                Pts.append(P)
                Atoms.append(Elem)
    return Pts, Atoms


def writeDesignToPOSCAR(D, filename, CommentLine=None, GS=None, BBox=None,
                        Elems=None, blnUseDirect=True):
    if GS is None:
        GS = 1.0
    if BBox is None:
        BBox = RectPrism.fromPointsBBox(D.Canvas.Points)
        BBox.scale(2.0)
    if Elems is None:
        Elems = D.NonVoidElems
    with open(filename, 'w') as outfile:
        outfile.write('{}\n'.format(CommentLine if CommentLine is not None else ''))
        outfile.write('{}\n'.format(GS))
        outfile.write('{} {} {}\n'.format(BBox.Vx[0] / GS, BBox.Vx[1] / GS, BBox.Vx[2] / GS))
        outfile.write('{} {} {}\n'.format(BBox.Vy[0] / GS, BBox.Vy[1] / GS, BBox.Vy[2] / GS))
        outfile.write('{} {} {}\n'.format(BBox.Vz[0] / GS, BBox.Vz[1] / GS, BBox.Vz[2] / GS))
        outfile.write('{}\n'.format(" ".join([Elem.Symbol for Elem in Elems])))
        outfile.write('{}\n'.format(" ".join([str(D.Contents.count(Elem)) for Elem in Elems])))
        outfile.write('{}\n'.format('Direct' if blnUseDirect else 'Cartesian'))
        for Elem in Elems:
            for i in range(len(D)):
                P = D.Canvas.Points[i] if blnUseDirect else BBox.getFractionalCoords(D.Canvas.Points[i])
                if D.Contents[i] == Elem:
                    outfile.write('{} {} {}\n'.format(P[0] / GS, P[1] / GS, P[2] / GS))



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Test for costing correlations of flash tank, fired heater,
and vessels (including platform and ladders, and distillation columns)

Created on Sept 25, 2020 by M. Zamarripa
"""
import pytest
# Import Pyomo libraries
import pyomo.environ as pyo
# Import IDAES core
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom
# Import Unit Model Modules
from idaes.generic_models.properties import iapws95
from idaes.core.util import get_solver
import idaes.core.util.unit_costing as cs
from idaes.power_generation.properties import FlueGasParameterBlock
from idaes.generic_models.unit_models.pressure_changer import (
    PressureChanger,
    ThermodynamicAssumption,
)
import idaes.core.util.unit_costing as costing
# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()

# -----------------------------------------------------------------------------


@pytest.mark.unit
def test_costing_FH_build():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.get_costing(year='2014', integer_n_units=True)
    m.fs.costing.CE_index = 550  # for testing only
    m.fs.unit = pyo.Block()
    m.fs.unit.heat_duty = pyo.Var(initialize=1e6,
                                  units=pyo.units.BTU/pyo.units.hr)
    m.fs.unit.pressure = pyo.Var(initialize=1e5,
                                 units=pyo.units.psi)
    m.fs.unit.costing = pyo.Block()
    m.fs.unit.heat_duty.fix(18390000)  # Btu/hr
    m.fs.unit.pressure.fix(700)  # psig

    cs.fired_heater_costing(m.fs.unit.costing,
                            fired_type='fuel',
                            Mat_factor='stain_steel',
                            ref_parameter_pressure=m.fs.unit.pressure,
                            ref_parameter_heat_duty=m.fs.unit.heat_duty)

    assert degrees_of_freedom(m) == 0
    # Check unit config arguments
    assert isinstance(m.fs.unit.costing.purchase_cost, pyo.Var)
    assert isinstance(m.fs.unit.costing.base_cost_per_unit, pyo.Var)


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_costing_FH_solve():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.get_costing()
    m.fs.costing.CE_index = 550  # for testing only
    m.fs.unit = pyo.Block()
    m.fs.unit.heat_duty = pyo.Var(initialize=1e6,
                                  units=pyo.units.BTU/pyo.units.hr)
    m.fs.unit.pressure = pyo.Var(initialize=1e5,
                                 units=pyo.units.psi)
    m.fs.unit.costing = pyo.Block()
    m.fs.unit.heat_duty.fix(18390000)  # Btu/hr
    m.fs.unit.pressure.fix(700)  # psig

    cs.fired_heater_costing(m.fs.unit.costing,
                            fired_type='fuel',
                            Mat_factor='stain_steel',
                            ref_parameter_pressure=m.fs.unit.pressure,
                            ref_parameter_heat_duty=m.fs.unit.heat_duty)

    assert degrees_of_freedom(m) == 0
    # Check unit config arguments
    assert isinstance(m.fs.unit.costing.purchase_cost, pyo.Var)
    assert isinstance(m.fs.unit.costing.base_cost_per_unit, pyo.Var)
    # initialize costing block
    costing.initialize(m.fs.unit.costing)
    assert (pytest.approx(pyo.value(m.fs.unit.costing.purchase_cost),
                          abs=1e-2) == 962795.521)
    results = solver.solve(m, tee=False)
    # Check for optimal solution
    assert results.solver.termination_condition == \
        pyo.TerminationCondition.optimal
    assert results.solver.status == pyo.SolverStatus.ok
    assert (pytest.approx(pyo.value(m.fs.unit.costing.purchase_cost),
                          abs=1e-2) == 962795.521)  # Example 22.1 Ref Book


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_costing_distillation_solve():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.get_costing()
    m.fs.costing.CE_index = 550
    # create a unit model and variables
    m.fs.unit = pyo.Block()
    m.fs.unit.heat_duty = pyo.Var(initialize=1e6,
                                  units=pyo.units.BTU/pyo.units.hr)
    m.fs.unit.pressure = pyo.Var(initialize=1e5,
                                 units=pyo.units.psi)
    m.fs.unit.diameter = pyo.Var(initialize=10,
                                 domain=pyo.NonNegativeReals,
                                 units=pyo.units.foot)
    m.fs.unit.length = pyo.Var(initialize=10,
                               domain=pyo.NonNegativeReals,
                               units=pyo.units.foot)
    # create costing block
    m.fs.unit.costing = pyo.Block()

    cs.vessel_costing(m.fs.unit.costing,
                      alignment='vertical',
                      weight_limit='option2',
                      L_D_range='option2',
                      PL=True,
                      plates=True,
                      number_tray=100,
                      ref_parameter_diameter=m.fs.unit.diameter,
                      ref_parameter_length=m.fs.unit.length)
    # pressure design and shell thickness from Example 22.13 Product and
    # Process Design Principless
    m.fs.unit.heat_duty.fix(18390000)  # Btu/hr
    m.fs.unit.pressure.fix(123+14.6959)  # psia
    # pressure design minimum thickness tp = 0.582 in
    # vessel is vertical + quite tall the tower is subject to wind load,
    # and earthquake. Assume wall thickness of 1.25 in.
    # The additional wall thickness at the bottom of the tower is 0.889 in
    # average thickness is 1.027, plus corrosion allowance of 1/8
    # 1.152 in, therefore steel plate thickness is 1.250 (ts)
    m.fs.unit.costing.shell_thickness.set_value(1.250)  # inches
    m.fs.unit.diameter.fix(10)  # ft
    m.fs.unit.length.fix(212)   # ft
    m.fs.unit.costing.number_trays.set_value(100)

    assert degrees_of_freedom(m) == 0
    # Check unit config arguments
    assert isinstance(m.fs.unit.costing.purchase_cost, pyo.Var)
    assert isinstance(m.fs.unit.costing.base_cost_per_unit, pyo.Var)

    results = solver.solve(m, tee=False)
    # Check for optimal solution
    assert results.solver.termination_condition == \
        pyo.TerminationCondition.optimal
    assert results.solver.status == pyo.SolverStatus.ok
    assert (pytest.approx(pyo.value(m.fs.unit.costing.base_cost),
                          abs=1e-2) == 636959.6929)  # Example 22.13 Ref Book
    assert (pytest.approx(pyo.value(m.fs.unit.costing.base_cost_platf_ladders),
                          abs=1e-2) == 97542.9005)  # Example 22.13 Ref Book
    assert (pytest.approx(pyo.value(m.fs.unit.costing.purchase_cost_trays),
                          abs=1e-2) == 293006.086)  # Example 22.13 Ref Book
    assert (pytest.approx(pyo.value(m.fs.unit.costing.purchase_cost),
                          abs=1e-2) == 1100958.9396)  # Example 22.13 Ref Book


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_blower_build_and_solve():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.properties = iapws95.Iapws95ParameterBlock()
    # Add property packages to flowsheet library
    m.fs.prop_fluegas = FlueGasParameterBlock()
    m.fs.unit = PressureChanger(default={
            "property_package": m.fs.prop_fluegas,
            "thermodynamic_assumption": ThermodynamicAssumption.isentropic,
            "compressor": True})

    # # FLUE GAS Inlet from Primary Superheater
    FGrate = 8516.27996  # mol/s
    # # Use FG molar composition to set component flow rates (baseline report)
    m.fs.unit.control_volume.properties_in[0].flow_mol_comp["H2O"].\
        fix(FGrate*8.69/100)
    m.fs.unit.control_volume.properties_in[0].flow_mol_comp["CO2"].\
        fix(FGrate*14.49/100)
    m.fs.unit.control_volume.properties_in[0].flow_mol_comp["N2"].\
        fix(FGrate*74.34/100)
    m.fs.unit.control_volume.properties_in[0].flow_mol_comp["O2"].\
        fix(FGrate*2.47/100)
    m.fs.unit.control_volume.properties_in[0].flow_mol_comp["NO"].\
        fix(FGrate*0.0006)
    m.fs.unit.control_volume.properties_in[0].flow_mol_comp["SO2"]\
        .fix(FGrate*0.002)
    m.fs.unit.control_volume.properties_in[0].temperature.fix(200.335)
    m.fs.unit.control_volume.properties_in[0].pressure.fix(99973.98)

    m.fs.unit.deltaP.fix(144790-99973.98)
    m.fs.unit.efficiency_isentropic.fix(0.9)

    m.fs.unit.get_costing(mover_type='fan')
    m.fs.unit.initialize()
    assert (pytest.approx(pyo.value(m.fs.unit.costing.purchase_cost),
                          abs=1e-2) == 272595.280)

    assert degrees_of_freedom(m) == 0
    # Check unit config arguments
    assert isinstance(m.fs.unit.costing.purchase_cost, pyo.Var)
    assert isinstance(m.fs.unit.costing.base_cost_per_unit, pyo.Var)
    results = solver.solve(m, tee=True)
    assert results.solver.termination_condition == \
        pyo.TerminationCondition.optimal
    assert results.solver.status == pyo.SolverStatus.ok
    assert (pytest.approx(pyo.value(m.fs.unit.costing.base_cost),
                          abs=1e-2) == 56026.447)
    assert (pytest.approx(pyo.value(m.fs.unit.costing.purchase_cost),
                          abs=1e-2) == 272595.280)


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_compressor_fan():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.properties = iapws95.Iapws95ParameterBlock()
    # Add property packages to flowsheet library
    m.fs.prop_fluegas = FlueGasParameterBlock()
    m.fs.unit = PressureChanger(default={
            "property_package": m.fs.prop_fluegas,
            "thermodynamic_assumption": ThermodynamicAssumption.isentropic,
            "compressor": True})

    # # FLUE GAS Inlet from Primary Superheater
    FGrate = 425.813998  # mol/s
    # # Use FG molar composition to set component flow rates (baseline report)
    m.fs.unit.control_volume.properties_in[0].flow_mol_comp["H2O"].\
        fix(FGrate*8.69/100)
    m.fs.unit.control_volume.properties_in[0].flow_mol_comp["CO2"].\
        fix(FGrate*14.49/100)
    m.fs.unit.control_volume.properties_in[0].flow_mol_comp["N2"].\
        fix(FGrate*74.34/100)
    m.fs.unit.control_volume.properties_in[0].flow_mol_comp["O2"].\
        fix(FGrate*2.47/100)
    m.fs.unit.control_volume.properties_in[0].flow_mol_comp["NO"].\
        fix(FGrate*0.0006)
    m.fs.unit.control_volume.properties_in[0].flow_mol_comp["SO2"]\
        .fix(FGrate*0.002)
    m.fs.unit.control_volume.properties_in[0].temperature.fix(200.335)
    m.fs.unit.control_volume.properties_in[0].pressure.fix(98658.6)

    m.fs.unit.deltaP.fix(101325-98658.6)
    m.fs.unit.efficiency_isentropic.fix(0.9)

    m.fs.unit.get_costing(mover_type='fan')

    m.fs.unit.initialize()
    # make sure costing initialized correctly
    assert (pytest.approx(pyo.value(m.fs.unit.costing.purchase_cost),
                          abs=1e-2) == 22106.9807)

    assert degrees_of_freedom(m) == 0
    # Check unit config arguments
    assert isinstance(m.fs.unit.costing.purchase_cost, pyo.Var)
    assert isinstance(m.fs.unit.costing.base_cost_per_unit, pyo.Var)
    results = solver.solve(m, tee=True)
    assert results.solver.termination_condition == \
        pyo.TerminationCondition.optimal
    assert results.solver.status == pyo.SolverStatus.ok
    assert (pytest.approx(pyo.value(m.fs.unit.costing.base_cost),
                          abs=1e-2) == 4543.6428)
    assert (pytest.approx(pyo.value(m.fs.unit.costing.purchase_cost),
                          abs=1e-2) == 22106.9807)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for Cubic equation of state methods

Author: Andrew Lee
"""
import pytest
from sys import modules
import os

from pyomo.environ import (ConcreteModel,
                           Expression,
                           ExternalFunction,
                           Param,
                           sqrt,
                           value,
                           Var,
                           units as pyunits)
from pyomo.core.expr.numeric_expr import ExternalFunctionExpression

from idaes.core import (declare_process_block_class,
                        LiquidPhase, VaporPhase, SolidPhase)
from idaes.generic_models.properties.core.eos.ceos import Cubic, CubicType
from idaes.generic_models.properties.core.generic.generic_property import (
        GenericParameterData)
from idaes.core.util.exceptions import PropertyNotSupportedError
from idaes.core.util.constants import Constants as const
from idaes import bin_directory


# Dummy method for property method calls
def dummy_call(b, j, T):
    return 42


# Dummy method to avoid errors when setting metadata dict
def set_metadata(b):
    pass


# Dummy methods for dummied submodules
class dummy_pe():
    def return_expression(b, *args):
        # Return a dummy expression for the constraint
        return b.temperature == 100


def phase_equil(b, *args):
    pass


@declare_process_block_class("DummyParameterBlock")
class DummyParameterData(GenericParameterData):
    def configure(self):
        self.configured = True

    def parameters(self):
        self.parameters_set = True


def define_state(b):
    # Add common variables
    b.pressure = Var(initialize=101325)
    b.temperature = Var(initialize=300)
    b.mole_frac_phase_comp = Var(b.params.phase_list,
                                 b.params.component_list,
                                 initialize=0.5)

    # Set mole fractions for each phase so they can be distinguished
    if "Sol" not in b.params.phase_list:
        b.mole_frac_phase_comp["Vap", "a"].value = 0.1
        b.mole_frac_phase_comp["Vap", "b"].value = 0.2
        b.mole_frac_phase_comp["Vap", "c"].value = 0.7
        b.mole_frac_phase_comp["Liq", "a"].value = 0.6
        b.mole_frac_phase_comp["Liq", "b"].value = 0.3
        b.mole_frac_phase_comp["Liq", "c"].value = 0.1


@pytest.fixture()
def m():
    m = ConcreteModel()

    # Dummy params block
    m.params = DummyParameterBlock(default={
                "components": {
                    "a": {"phase_equilibrium_form": {("Vap", "Liq"): dummy_pe},
                          "parameter_data": {
                              "omega": 0.1,
                              "pressure_crit": 1e5,
                              "temperature_crit": 100}},
                    "b": {"phase_equilibrium_form": {("Vap", "Liq"): dummy_pe},
                          "parameter_data": {
                              "omega": 0.2,
                              "pressure_crit": 2e5,
                              "temperature_crit": 200}},
                    "c": {"phase_equilibrium_form": {("Vap", "Liq"): dummy_pe},
                          "parameter_data": {
                              "omega": 0.3,
                              "pressure_crit": 3e5,
                              "temperature_crit": 300}}},
                "phases": {
                    "Vap": {"type": VaporPhase,
                            "equation_of_state": Cubic,
                            "equation_of_state_options": {
                                "type": CubicType.PR}},
                    "Liq": {"type": LiquidPhase,
                            "equation_of_state": Cubic,
                            "equation_of_state_options": {
                                "type": CubicType.PR}}},
                "base_units": {"time": pyunits.s,
                               "length": pyunits.m,
                               "mass": pyunits.kg,
                               "amount": pyunits.mol,
                               "temperature": pyunits.K},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "phases_in_equilibrium": [("Vap", "Liq")],
                "phase_equilibrium_state": {("Vap", "Liq"): modules[__name__]},
                "parameter_data": {"PR_kappa": {("a", "a"): 0.000,
                                                ("a", "b"): 0.000,
                                                ("a", "c"): 0.000,
                                                ("b", "a"): 0.000,
                                                ("b", "b"): 0.000,
                                                ("b", "c"): 0.000,
                                                ("c", "a"): 0.000,
                                                ("c", "b"): 0.000,
                                                ("c", "c"): 0.000}}})

    m.props = m.params.state_block_class(
        [1], default={"defined_state": False,
                      "parameters": m.params,
                      "has_phase_equilibrium": True})

    # Set a distinct value for _teq so it can be distinguished from temperature
    m.props[1]._teq[("Vap", "Liq")].value = 100

    return m


@pytest.fixture()
def m_sol():
    m = ConcreteModel()

    # Dummy params block
    m.params = DummyParameterBlock(default={
                "components": {
                    "a": {"phase_equilibrium_form": {("Sol", "Liq"): dummy_pe},
                          "parameter_data": {
                              "omega": 0.1,
                              "pressure_crit": 1e5,
                              "temperature_crit": 100}},
                    "b": {"phase_equilibrium_form": {("Sol", "Liq"): dummy_pe},
                          "parameter_data": {
                              "omega": 0.2,
                              "pressure_crit": 2e5,
                              "temperature_crit": 200}},
                    "c": {"phase_equilibrium_form": {("Sol", "Liq"): dummy_pe},
                          "parameter_data": {
                              "omega": 0.3,
                              "pressure_crit": 3e5,
                              "temperature_crit": 300}}},
                "phases": {
                    "Sol": {"type": SolidPhase,
                            "equation_of_state": Cubic,
                            "equation_of_state_options": {
                                "type": CubicType.PR}},
                    "Liq": {"type": LiquidPhase,
                            "equation_of_state": Cubic,
                            "equation_of_state_options": {
                                "type": CubicType.PR}}},
                "base_units": {"time": pyunits.s,
                               "length": pyunits.m,
                               "mass": pyunits.kg,
                               "amount": pyunits.mol,
                               "temperature": pyunits.K},
                "state_definition": modules[__name__],
                "pressure_ref": 1e5,
                "temperature_ref": 300,
                "phases_in_equilibrium": [("Sol", "Liq")],
                "phase_equilibrium_state": {("Sol", "Liq"): modules[__name__]},
                "parameter_data": {"PR_kappa": {("a", "a"): 0.000,
                                                ("a", "b"): 0.000,
                                                ("a", "c"): 0.000,
                                                ("b", "a"): 0.000,
                                                ("b", "b"): 0.000,
                                                ("b", "c"): 0.000,
                                                ("c", "a"): 0.000,
                                                ("c", "b"): 0.000,
                                                ("c", "c"): 0.000}}})

    m.props = m.params.state_block_class(
        [1], default={"defined_state": False,
                      "parameters": m.params,
                      "has_phase_equilibrium": True})

    return m


# Expected values for A and B
Al = 0.0419062
Bl = 0.0262770
Av = 0.1180005
Bv = 0.0262769
Zl = 0.9870125
Zv = 0.9067390

Al_eq = 0.8325554
Bl_eq = 0.0788309
Av_eq = 1.9805792
Bv_eq = 0.0788309

Hl = -212.211
Hv = -779.305
Ul = Hl - 8.314*300*(Zl-1)
Uv = Hv - 8.314*300*(Zv-1)

# Set path to root finder .so file
_so = os.path.join(bin_directory, "cubic_roots.so")
f_Zl = ExternalFunction(library=_so, function="ceos_z_liq")
f_Zv = ExternalFunction(library=_so, function="ceos_z_vap")


@pytest.mark.unit
def test_common(m):
    # Test cubic components
    assert isinstance(m.props[1].PR_fw, Expression)
    assert len(m.props[1].PR_fw) == len(m.params.component_list)
    for i in m.params.component_list:
        omega = m.params.get_component(i).omega
        assert value(m.props[1].PR_fw[i]) == value(
            0.37464 + 1.54226*omega - 0.26992*omega**2)

    assert isinstance(m.props[1].PR_a, Expression)
    assert len(m.props[1].PR_a) == len(m.params.component_list)
    for i in m.params.component_list:
        Tc = m.params.get_component(i).temperature_crit
        Pc = m.params.get_component(i).pressure_crit
        assert pytest.approx(value(m.props[1].PR_a[i]),
                             rel=1e-5) == value(
            0.45724*((const.gas_constant*Tc)**2/Pc) *
            (((1+m.props[1].PR_fw[i] *
               (1-sqrt(m.props[1].temperature/Tc)))**2)))

    assert isinstance(m.props[1].PR_b, Expression)
    assert len(m.props[1].PR_b) == len(m.params.component_list)
    for i in m.params.component_list:
        Tc = m.params.get_component(i).temperature_crit
        Pc = m.params.get_component(i).pressure_crit
        assert value(m.props[1].PR_b[i]) == value(
            0.07780*const.gas_constant*Tc/Pc)

    assert isinstance(m.props[1].PR_am, Expression)
    assert len(m.props[1].PR_am) == len(m.params.phase_list)
    for p in m.params.phase_list:
        assert pytest.approx(value(m.props[1].PR_am[p]),
                             rel=1e-5) == value(
            sum(sum(m.props[1].mole_frac_phase_comp[p, i] *
                    m.props[1].mole_frac_phase_comp[p, j] *
                    sqrt(m.props[1].PR_a[i]*m.props[1].PR_a[j]) *
                    (1-m.params.PR_kappa[i, j])
                    for j in m.params.component_list)
                for i in m.params.component_list))

    assert isinstance(m.props[1].PR_bm, Expression)
    assert len(m.props[1].PR_bm) == len(m.params.phase_list)
    for p in m.params.phase_list:
        assert pytest.approx(value(m.props[1].PR_bm[p]),
                             rel=1e-5) == value(
            sum(m.props[1].mole_frac_phase_comp[p, i] *
                m.props[1].PR_b[i]
                for i in m.props[1].params.component_list))

    assert isinstance(m.props[1].PR_A, Expression)
    assert len(m.props[1].PR_A) == len(m.params.phase_list)
    for p in m.params.phase_list:
        assert pytest.approx(value(m.props[1].PR_A[p]),
                             rel=1e-5) == value(
                m.props[1].PR_am[p]*m.props[1].pressure /
                (const.gas_constant*m.props[1].temperature)**2)
    assert pytest.approx(value(m.props[1].PR_A["Liq"]), rel=1e-5) == Al
    assert pytest.approx(value(m.props[1].PR_A["Vap"]), rel=1e-5) == Av

    assert isinstance(m.props[1].PR_B, Expression)
    assert len(m.props[1].PR_B) == len(m.params.phase_list)
    for p in m.params.phase_list:
        assert pytest.approx(value(m.props[1].PR_B[p]),
                             rel=1e-5) == value(
                m.props[1].PR_bm[p]*m.props[1].pressure /
                (const.gas_constant*m.props[1].temperature))
    assert pytest.approx(value(m.props[1].PR_B["Liq"]), rel=1e-5) == Bl
    assert pytest.approx(value(m.props[1].PR_B["Vap"]), rel=1e-5) == Bv

    assert isinstance(m.props[1].PR_delta, Expression)
    assert len(m.props[1].PR_delta) == 6
    for p in m.params.phase_list:
        for i in m.params.component_list:
            assert pytest.approx(value(m.props[1].PR_delta[p, i]),
                                 rel=1e-5) == value(
                2*sqrt(m.props[1].PR_a[i])/m.props[1].PR_am[p] *
                sum(m.props[1].mole_frac_phase_comp[p, j] *
                    sqrt(m.props[1].PR_a[j]) *
                    (1-m.params.PR_kappa[i, j])
                    for j in m.params.component_list))

    assert isinstance(m.props[1].PR_dadT, Expression)
    assert len(m.props[1].PR_dadT) == len(m.params.phase_list)
    for p in m.params.phase_list:
        assert pytest.approx(value(m.props[1].PR_dadT[p]),
                             rel=1e-5) == value(
            -((const.gas_constant/2)*sqrt(0.45724) *
              sum(sum(m.props[1].mole_frac_phase_comp[p, i] *
                      m.props[1].mole_frac_phase_comp[p, j] *
                      (1-m.params.PR_kappa[i, j]) *
                      (m.props[1].PR_fw[j] *
                       sqrt(m.props[1].PR_a[i] *
                            m.params.get_component(j).temperature_crit /
                            m.params.get_component(j).pressure_crit) +
                       m.props[1].PR_fw[i] *
                       sqrt(m.props[1].PR_a[j] *
                            m.params.get_component(i).temperature_crit /
                            m.params.get_component(i).pressure_crit))
                      for j in m.params.component_list)
                  for i in m.params.component_list) /
              sqrt(m.props[1].temperature)))

    # Test equilibrium state Expressions
    assert isinstance(m.props[1]._PR_a_eq, Expression)
    assert len(m.props[1]._PR_a_eq) == len(m.params.component_list)
    for i in m.props[1]._PR_a_eq:
        Tc = m.params.get_component(i[2]).temperature_crit
        Pc = m.params.get_component(i[2]).pressure_crit
        assert pytest.approx(value(m.props[1]._PR_a_eq[i]),
                             rel=1e-5) == value(
            0.45724*((const.gas_constant*Tc)**2/Pc) *
            (((1+m.props[1].PR_fw[i[2]] *
               (1-sqrt(m.props[1]._teq[i[0], i[1]]/Tc)))**2)))

    assert isinstance(m.props[1]._PR_am_eq, Expression)
    assert len(m.props[1]._PR_am_eq) == len(m.params.phase_list)
    for idx in m.props[1]._PR_am_eq:
        assert pytest.approx(value(m.props[1]._PR_am_eq[idx]),
                             rel=1e-5) == value(
            sum(sum(m.props[1].mole_frac_phase_comp[idx[2], i] *
                    m.props[1].mole_frac_phase_comp[idx[2], j] *
                    sqrt(m.props[1]._PR_a_eq[idx[0], idx[1], i] *
                         m.props[1]._PR_a_eq[idx[0], idx[1], j]) *
                    (1-m.params.PR_kappa[i, j])
                    for j in m.params.component_list)
                for i in m.params.component_list))

    assert isinstance(m.props[1]._PR_A_eq, Expression)
    assert len(m.props[1]._PR_A_eq) == len(m.params.phase_list)
    for i in m.props[1]._PR_A_eq:
        assert pytest.approx(value(m.props[1]._PR_A_eq[i]),
                             rel=1e-5) == value(
                    m.props[1]._PR_am_eq[i] *
                    m.props[1].pressure /
                    (const.gas_constant *
                     m.props[1]._teq[i[0], i[1]])**2)
    assert pytest.approx(value(m.props[1]._PR_A_eq[("Vap", "Liq"), "Liq"]),
                         rel=1e-5) == Al_eq
    assert pytest.approx(value(m.props[1]._PR_A_eq[("Vap", "Liq"), "Vap"]),
                         rel=1e-5) == Av_eq

    assert isinstance(m.props[1]._PR_B_eq, Expression)
    assert len(m.props[1]._PR_B_eq) == len(m.params.phase_list)
    for i in m.props[1]._PR_B_eq:
        assert pytest.approx(value(m.props[1]._PR_B_eq[i]),
                             rel=1e-5) == value(
                    m.props[1].PR_bm[i[2]] *
                    m.props[1].pressure /
                    (const.gas_constant *
                     m.props[1]._teq[i[0], i[1]]))
    assert pytest.approx(value(m.props[1]._PR_B_eq[("Vap", "Liq"), "Liq"]),
                         rel=1e-5) == Bl_eq
    assert pytest.approx(value(m.props[1]._PR_B_eq[("Vap", "Liq"), "Vap"]),
                         rel=1e-5) == Bv_eq

    assert isinstance(m.props[1]._PR_delta_eq, Expression)
    assert len(m.props[1]._PR_delta_eq) == 6
    for idx in m.props[1]._PR_delta_eq:
        assert pytest.approx(value(m.props[1]._PR_delta_eq[idx]),
                             rel=1e-5) == value(
            2*sqrt(m.props[1]._PR_a_eq[idx[0], idx[1], idx[3]]) /
            m.props[1]._PR_am_eq[idx[0], idx[1], idx[2]] *
            sum(m.props[1].mole_frac_phase_comp[idx[2], j] *
                sqrt(m.props[1]._PR_a_eq[idx[0], idx[1], j]) *
                (1-m.params.PR_kappa[idx[3], j])
                for j in m.params.component_list))

    # Check for external function components
    assert isinstance(m.props[1]._PR_ext_func_param, Param)
    assert m.props[1]._PR_ext_func_param.value == 0  # 0 == PR
    assert isinstance(m.props[1]._PR_proc_Z_liq, ExternalFunction)
    assert isinstance(m.props[1]._PR_proc_Z_vap, ExternalFunction)


@pytest.mark.unit
def test_compress_fact_phase_Liq(m):
    assert isinstance(Cubic.compress_fact_phase(m.props[1], "Liq"),
                      ExternalFunctionExpression)
    assert pytest.approx(value(f_Zl(0, Al, Bl)), rel=1e-5) == value(
        Cubic.compress_fact_phase(m.props[1], "Liq"))
    assert pytest.approx(value(
        Cubic.compress_fact_phase(m.props[1], "Liq")), rel=1e-5) == Zl


@pytest.mark.unit
def test_compress_fact_phase_Vap(m):
    assert isinstance(Cubic.compress_fact_phase(m.props[1], "Vap"),
                      ExternalFunctionExpression)
    assert pytest.approx(value(f_Zv(0, Av, Bv)), rel=1e-5) == value(
        Cubic.compress_fact_phase(m.props[1], "Vap"))
    assert pytest.approx(value(
        Cubic.compress_fact_phase(m.props[1], "Vap")), rel=1e-5) == Zv


@pytest.mark.unit
def test_compress_fact_phase_invalid_phase(m_sol):
    with pytest.raises(PropertyNotSupportedError):
        Cubic.compress_fact_phase(m_sol.props[1], "Sol")


@pytest.mark.unit
def test_dens_mass_phase(m):
    m.props[1].dens_mol_phase = Var(m.params.phase_list)
    m.props[1].mw_phase = Var(m.params.phase_list)

    for p in m.params.phase_list:
        assert str(Cubic.dens_mass_phase(m.props[1], p)) == str(
                m.props[1].dens_mol_phase[p]*m.props[1].mw_phase[p])


@pytest.mark.unit
def test_dens_mol_phase(m):
    assert value(Cubic.dens_mol_phase(m.props[1], "Vap")) == pytest.approx(
            44.800, rel=1e-3)
    assert value(Cubic.dens_mol_phase(m.props[1], "Liq")) == pytest.approx(
            41.157, rel=1e-3)


@pytest.mark.unit
def test_dens_mol_phase_invalid_phase(m_sol):
    with pytest.raises(PropertyNotSupportedError):
        Cubic.dens_mol_phase(m_sol.props[1], "Sol")


@pytest.mark.unit
def test_energy_internal_mol_phase(m):
    for j in m.params.component_list:
        m.params.config.include_enthalpy_of_formation = False
        m.params.get_component(j).config.enth_mol_liq_comp = dummy_call
        m.params.get_component(j).config.enth_mol_ig_comp = dummy_call

    m.props[1].energy_internal_mol_phase_comp = Var(
        m.params.phase_list, m.params.component_list, initialize=1)

    assert pytest.approx(Uv, rel=1e-4) == value(
        Cubic.energy_internal_mol_phase(m.props[1], "Vap"))
    assert pytest.approx(Ul, rel=1e-4) == value(
        Cubic.energy_internal_mol_phase(m.props[1], "Liq"))


@pytest.mark.unit
def test_energy_internal_mol_phase_comp(m):
    for j in m.params.component_list:
        m.params.config.include_enthalpy_of_formation = False
        m.params.get_component(j).config.enth_mol_liq_comp = dummy_call
        m.params.get_component(j).config.enth_mol_ig_comp = dummy_call

        assert pytest.approx(Ul, rel=1e-4) == value(
            Cubic.energy_internal_mol_phase_comp(m.props[1], "Liq", j))
        assert pytest.approx(Uv, rel=1e-4) == value(
            Cubic.energy_internal_mol_phase_comp(m.props[1], "Vap", j))


@pytest.mark.unit
def test_energy_internal_mol_phase_invalid_phase(m_sol):
    with pytest.raises(PropertyNotSupportedError):
        Cubic.energy_internal_mol_phase_comp(m_sol.props[1], "Sol", "foo")


@pytest.mark.unit
def test_enth_mol_phase(m):
    for j in m.params.component_list:
        m.params.get_component(j).config.enth_mol_liq_comp = dummy_call
        m.params.get_component(j).config.enth_mol_ig_comp = dummy_call

    m.props[1].enth_mol_phase_comp = Var(m.params.phase_list,
                                         m.params.component_list,
                                         initialize=1)

    assert pytest.approx(value(
        Cubic.enth_mol_phase(m.props[1], "Vap")), rel=1e-5) == Hv
    assert pytest.approx(value(
        Cubic.enth_mol_phase(m.props[1], "Liq")), rel=1e-5) == Hl


@pytest.mark.unit
def test_enth_mol_phase_comp(m):
    for j in m.params.component_list:
        m.params.get_component(j).config.enth_mol_liq_comp = dummy_call
        m.params.get_component(j).config.enth_mol_ig_comp = dummy_call

        assert pytest.approx(Hl, rel=1e-5) == value(
            Cubic.enth_mol_phase_comp(m.props[1], "Liq", j))
        assert pytest.approx(Hv, rel=1e-5) == value(
            Cubic.enth_mol_phase_comp(m.props[1], "Vap", j))


@pytest.mark.unit
def test_enth_mol_phase_invalid_phase(m_sol):
    with pytest.raises(PropertyNotSupportedError):
        Cubic.enth_mol_phase_comp(m_sol.props[1], "Sol", "foo")


@pytest.mark.unit
def test_entr_mol_phase(m):
    for j in m.params.component_list:
        m.params.get_component(j).config.entr_mol_liq_comp = dummy_call
        m.params.get_component(j).config.entr_mol_ig_comp = dummy_call

    m.props[1].entr_mol_phase_comp = Var(m.params.phase_list,
                                         m.params.component_list,
                                         initialize=1)

    assert pytest.approx(value(
        Cubic.entr_mol_phase(m.props[1], "Vap")), rel=1e-5) == 39.9218
    assert pytest.approx(value(
        Cubic.entr_mol_phase(m.props[1], "Liq")), rel=1e-5) == 41.1621


@pytest.mark.unit
def test_entr_mol_phase_comp(m):
    entr = {("Liq", "a"): 45.4093,
            ("Vap", "a"): 59.0666,
            ("Liq", "b"): 51.1725,
            ("Vap", "b"): 53.3035,
            ("Liq", "c"): 60.3069,
            ("Vap", "c"): 42.8875}
    for j in m.params.component_list:
        m.params.get_component(j).config.entr_mol_liq_comp = dummy_call
        m.params.get_component(j).config.entr_mol_ig_comp = dummy_call

        assert pytest.approx(entr[("Liq", j)], rel=1e-5) == value(
            Cubic.entr_mol_phase_comp(m.props[1], "Liq", j))
        assert pytest.approx(entr[("Vap", j)], rel=1e-5) == value(
            Cubic.entr_mol_phase_comp(m.props[1], "Vap", j))


@pytest.mark.unit
def test_entr_mol_phase_invalid_phase(m_sol):
    with pytest.raises(PropertyNotSupportedError):
        Cubic.entr_mol_phase_comp(m_sol.props[1], "Sol", "foo")


@pytest.mark.component
def test_fug_phase_comp(m):
    for p in m.params.phase_list:
        for j in m.params.component_list:
            assert str(Cubic.fug_phase_comp(
                            m.props[1], p, j)) == str(
                m.props[1].mole_frac_phase_comp[p, j] *
                m.props[1].pressure *
                m.props[1].fug_coeff_phase_comp[p, j])


@pytest.mark.unit
def test_fug_phase_comp_invalid_phase(m_sol):
    with pytest.raises(PropertyNotSupportedError):
        Cubic.fug_phase_comp(m_sol.props[1], "Sol", "foo")


@pytest.mark.component
def test_fug_phase_comp_eq(m):
    for p in m.params.phase_list:
        for j in m.params.component_list:
            assert str(Cubic.fug_phase_comp_eq(
                            m.props[1], p, j, ("Vap", "Liq"))) == str(
                m.props[1].mole_frac_phase_comp[p, j] *
                m.props[1].pressure *
                Cubic.fug_coeff_phase_comp_eq(
                    m.props[1], p, j, ("Vap", "Liq")))


@pytest.mark.unit
def test_fug_phase_comp_eq_invalid_phase(m_sol):
    with pytest.raises(PropertyNotSupportedError):
        Cubic.fug_phase_comp_eq(m_sol.props[1], "Sol", "foo", ("Vap", "Liq"))


@pytest.mark.unit
def test_fug_coeff_phase_comp_Liq(m):
    assert pytest.approx(1.01213, rel=1e-5) == value(
        Cubic.fug_coeff_phase_comp(m.props[1], "Liq", "a"))
    assert pytest.approx(0.95919, rel=1e-5) == value(
        Cubic.fug_coeff_phase_comp(m.props[1], "Liq", "b"))
    assert pytest.approx(0.91356, rel=1e-5) == value(
        Cubic.fug_coeff_phase_comp(m.props[1], "Liq", "c"))


@pytest.mark.unit
def test_fug_coeff_phase_comp_Vap(m):
    assert pytest.approx(1.05952, rel=1e-5) == value(
            Cubic.fug_coeff_phase_comp(m.props[1], "Vap", "a"))
    assert pytest.approx(0.96070, rel=1e-5) == value(
        Cubic.fug_coeff_phase_comp(m.props[1], "Vap", "b"))
    assert pytest.approx(0.87903, rel=1e-5) == value(
        Cubic.fug_coeff_phase_comp(m.props[1], "Vap", "c"))


@pytest.mark.unit
def test_fug_coeff_phase_comp_invalid_phase(m_sol):
    with pytest.raises(PropertyNotSupportedError):
        Cubic.fug_coeff_phase_comp(m_sol.props[1], "Sol", "foo")


@pytest.mark.unit
def test_fug_coeff_phase_comp_eq_Liq(m):
    assert pytest.approx(1.22431, rel=1e-5) == value(
        Cubic.fug_coeff_phase_comp_eq(m.props[1], "Liq", "a", ("Vap", "Liq")))
    assert pytest.approx(0.0049604, rel=1e-5) == value(
        Cubic.fug_coeff_phase_comp_eq(m.props[1], "Liq", "b", ("Vap", "Liq")))
    assert pytest.approx(3.19128e-05, rel=1e-5) == value(
        Cubic.fug_coeff_phase_comp_eq(m.props[1], "Liq", "c", ("Vap", "Liq")))


@pytest.mark.unit
def test_fug_coeff_phase_comp_eq_Vap(m):
    assert pytest.approx(86.9140, rel=1e-5) == value(
        Cubic.fug_coeff_phase_comp_eq(m.props[1], "Vap", "a", ("Vap", "Liq")))
    assert pytest.approx(0.0049878, rel=1e-5) == value(
        Cubic.fug_coeff_phase_comp_eq(m.props[1], "Vap", "b", ("Vap", "Liq")))
    assert pytest.approx(6.49776e-07, rel=1e-5) == value(
        Cubic.fug_coeff_phase_comp_eq(m.props[1], "Vap", "c", ("Vap", "Liq")))


@pytest.mark.unit
def test_fug_coeff_phase_comp_eq_invalid_phase(m_sol):
    with pytest.raises(PropertyNotSupportedError):
        Cubic.fug_coeff_phase_comp_eq(
            m_sol.props[1], "Sol", "foo", ("Vap", "Liq"))


@pytest.mark.unit
def test_gibbs_mol_phase(m):
    m.props[1].enth_mol_phase = Var(m.params.phase_list)
    m.props[1].entr_mol_phase = Var(m.params.phase_list)

    for p in m.params.phase_list:
        assert str(Cubic.gibbs_mol_phase(m.props[1], p)) == str(
            m.props[1].enth_mol_phase[p] -
            m.props[1].entr_mol_phase[p]*m.props[1].temperature)


@pytest.mark.unit
def test_gibbs_mol_phase_comp(m):
    m.props[1].enth_mol_phase_comp = Var(m.params.phase_list,
                                         m.params.component_list)
    m.props[1].entr_mol_phase_comp = Var(m.params.phase_list,
                                         m.params.component_list)

    for p in m.params.phase_list:
        for j in m.params.component_list:
            assert str(Cubic.gibbs_mol_phase_comp(m.props[1], p, j)) == str(
                    m.props[1].enth_mol_phase_comp[p, j] -
                    m.props[1].entr_mol_phase_comp[p, j] *
                    m.props[1].temperature)
